import sys
import os
import logging

logger = logging.Logger("plugin_registry")

ESCAPE_DECODING = 'unicode_escape'

sys.path.append(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__))))
from notest import generators
from notest import validators
from notest import operations


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def register_extensions(modules):
    """ Import the modules and register their respective extensions """
    if isinstance(modules, str):  # Catch supplying just a string arg
        modules = [modules]
    for ext in modules:
        # Get the package prefix and final module name
        segments = ext.split('.')
        module = segments.pop()
        package = '.'.join(segments)
        # Necessary to get the root module back
        module = __import__(ext, globals(), locals(), package)

        # Extensions are registered by applying a register function to sets of
        # registry name/function pairs inside an object
        extension_applies = {
            'VALIDATORS': validators.register_validator,
            'COMPARATORS': validators.register_comparator,
            'VALIDATOR_TESTS': validators.register_test,
            'EXTRACTORS': validators.register_extractor,
            'GENERATORS': generators.register_generator,
            'OPERATIONS': operations.register_operations
        }

        has_registry = False
        for registry_name, register_function in extension_applies.items():
            if hasattr(module, registry_name):
                registry = getattr(module, registry_name)
                for key, val in registry.items():
                    register_function(key, val)
                    logger.info("Register {} {} to module {}".format(
                        registry_name.lower(), key, ext
                    ))
                if registry:
                    has_registry = True

        if not has_registry:
            raise ImportError(
                "Extension to register did not contain any registries: {0}".format(
                    ext))


def auto_load_ext(ext_dir=os.path.join(BASE_DIR, "ext")):
    if not os.path.isdir(ext_dir):
        return
    sys.path.append(ext_dir)
    for fname in os.listdir(ext_dir):
        path = os.path.join(ext_dir, fname)
        if os.path.isfile(path) and fname.endswith(".py") and \
                fname != "__init__.py":
            module_name = fname[:-3].split(os.path.sep)
            try:
                register_extensions(".".join(module_name))
            except ImportError as e:
                logger.error(str(e))
    print()


auto_load_ext(ext_dir=os.path.join(BASE_DIR, "ext"))



from notest.lib.utils import templated_var

'''
- operation:
    - type: "print_operation"
    - print: 'hello world'
'''

def print_operation(config, context=None):
    print("Run print_operation")
    assert isinstance(config, dict)
    assert "print" in config
    pr = config['print']
    pr = templated_var(pr, context)
    print(pr)
    print('')


OPERATIONS = {'print_operation': print_operation}




import string
import yaml
import json


def templated_string(src, context=None):
    if not context:
        return src
    if not isinstance(context, dict):
        context = context.get_values()
    src = string.Template(src).safe_substitute(context)
    return src


def templated_var(data, context=None):
    if not context:
        return data
    if isinstance(data, str) and data.find("$") > -1:
        res = templated_string(data, context)
        return res
    elif isinstance(data, dict) and len(data) == 1 and "template" in data:
        src = data["template"]
        res = templated_string(src, context)
        return res
    else:
        return data


def read_file(path):
    """ Read an input into a file, doing necessary conversions around relative path handling """
    with open(path, "r") as f:
        string = f.read()
        f.close()
    return string


def read_test_file(path):
    """ Read test file at 'path' in yaml/yml or json"""
    content = read_file(path)
    try:
        teststruct = yaml.safe_load(content)
    except:
        teststruct = json.loads(content)
    return teststruct



class HttpResponse:
    def __init__(self, body=None, headers=None, status_code=None,
                 reason=None, cookies=None, requests_response=None):
        if requests_response is not None:
            self.body = requests_response.content
            requests_headers = requests_response.headers
            self.headers = [i for i in requests_headers.lower_items()]
            # self.headers = {k: v for k, v in requests_headers.lower_items()}
            self.status_code = requests_response.status_code
            self.reason = requests_response.reason
            self.cookies = requests_response.cookies
        else:
            self.body = body
            self.headers = headers
            self.status_code = status_code
            self.reason = reason
            self.cookies = cookies




# DEFAULT_TIMEOUT = 10  # Seconds
# HTTP_CLIENT = "requests"    # requests or pycurl
#
# if HTTP_CLIENT == "pycurl":
#     try:
#         import pycurl
#     except:
#         print("pycurl is not installed successfully, change to requests")
#         HTTP_CLIENT = "requests"
#
# if HTTP_CLIENT == "pycurl":
#     from notest.clients.pycurl_client import PyCurlClient as HttpClient
# elif HTTP_CLIENT == "requests":
#     from notest.clients.requests_client import RequestsClient as HttpClient
# else:
#     raise Exception("Unknown Http Client Type: {}".format(HTTP_CLIENT))


def get_client_class(client_type="requests"):
    if client_type == "pycurl":
        from notest.clients.pycurl_client import PyCurlClient as HttpClient
        return HttpClient
    elif client_type == "requests":
        from notest.clients.requests_client import RequestsClient as HttpClient
        return HttpClient
    else:
        raise Exception("Unknown Client Type: {}".format(client_type))



import pycurl
import os
import sys
import copy
from io import BytesIO
from .http_auth_type import HttpAuthType
from .http_response import HttpResponse


base_dir = os.path.abspath(os.path.dirname(__file__))
libcurl_crt_file = os.path.join(base_dir, "..", "..", "tools", "curl-ca-bundle.crt")


DEFAULT_TIMEOUT = 10  # Seconds
# Map HTTP method names to curl methods
# Kind of obnoxious that it works this way...
HTTP_METHODS = {'GET': pycurl.HTTPGET,
                'PUT': pycurl.UPLOAD,
                'PATCH': pycurl.POSTFIELDS,
                'POST': pycurl.POST,
                'HEAD': "",
                'DELETE': 'DELETE'}


HttpAuthType_Map = {
    HttpAuthType.HTTP_AUTH_BASIC: pycurl.HTTPAUTH_BASIC
}


class PyCurlClient:
    def __init__(self, handler=None):
        if not handler or not isinstance(handler, pycurl.Curl):
            self.handler = pycurl.Curl()
        else:
            self.handler = handler
        self.response = None

    def get_handler(self):
        return self.handler

    @staticmethod
    def close_handler(handler):
        if handler:
            handler.close()

    def close(self):
        if self.handler:
            self.handler.close()

    def send_request(self, test_obj, timeout=DEFAULT_TIMEOUT, context=None,
                     handler=None, ssl_insecure=True, verbose=False):
        """ Create and mostly configure a curl object for test, reusing existing if possible """

        if handler:
            curl = handler

            try:  # Check the curl handle isn't closed, and reuse it if possible
                curl.getinfo(curl.HTTP_CODE)
                # Below clears the cookies & curl options for clean run
                # But retains the DNS cache and connection pool
                curl.reset()
                curl.setopt(curl.COOKIELIST, "ALL")
            except pycurl.error:
                curl = pycurl.Curl()

        else:
            curl = self.handler

        curl.setopt(curl.URL, str(test_obj.url))
        curl.setopt(curl.TIMEOUT, timeout)

        is_unicoded = False
        _body = test_obj.body
        if isinstance(_body, str):  # Encode unicode
            _body = _body.encode('UTF-8')
            is_unicoded = True

        # Set read function for post/put bodies
        if _body and len(_body) > 0:
            curl.setopt(curl.READFUNCTION, BytesIO(_body).read)

        if test_obj.auth_username and test_obj.auth_password:
            auth_username = test_obj.auth_username
            auth_password = test_obj.auth_password
            if isinstance(auth_username, str):
                auth_username = auth_username.encode()
            if isinstance(auth_password, str):
                auth_password = auth_password.encode()
            curl.setopt(pycurl.USERPWD, auth_username + b':' + auth_password)
            if test_obj.auth_type:
                auth_type = HttpAuthType_Map[test_obj.auth_type]
                curl.setopt(pycurl.HTTPAUTH, auth_type)

        if test_obj.method == u'POST':
            curl.setopt(pycurl.POST, 1)
            # Required for some servers
            if _body is not None:
                curl.setopt(pycurl.POSTFIELDSIZE, len(_body))
            else:
                curl.setopt(pycurl.POSTFIELDSIZE, 0)
        elif test_obj.method == u'PUT':
            curl.setopt(pycurl.UPLOAD, 1)
            # Required for some servers
            if _body is not None:
                curl.setopt(pycurl.INFILESIZE, len(_body))
            else:
                curl.setopt(pycurl.INFILESIZE, 0)
        elif test_obj.method == u'PATCH':
            curl.setopt(curl.POSTFIELDS, _body)
            curl.setopt(curl.CUSTOMREQUEST, 'PATCH')
            # Required for some servers
            # I wonder: how compatible will this be?  It worked with Django but feels iffy.
            if _body is not None:
                curl.setopt(pycurl.INFILESIZE, len(_body))
            else:
                curl.setopt(pycurl.INFILESIZE, 0)
        elif test_obj.method == u'DELETE':
            curl.setopt(curl.CUSTOMREQUEST, 'DELETE')
            if _body is not None:
                curl.setopt(pycurl.POSTFIELDS, _body)
                curl.setopt(pycurl.POSTFIELDSIZE, len(_body))
        elif test_obj.method == u'HEAD':
            curl.setopt(curl.NOBODY, 1)
            curl.setopt(curl.CUSTOMREQUEST, 'HEAD')
        elif test_obj.method and test_obj.method.upper() != 'GET':  # Alternate HTTP methods
            curl.setopt(curl.CUSTOMREQUEST, test_obj.method.upper())
            if _body is not None:
                curl.setopt(pycurl.POSTFIELDS, _body)
                curl.setopt(pycurl.POSTFIELDSIZE, len(_body))

        # Template headers as needed and convert headers dictionary to list of header entries
        head = test_obj.get_headers(context=context)
        head = copy.copy(head)  # We're going to mutate it, need to copy

        # Set charset if doing unicode conversion and not set explicitly
        # TESTME
        if is_unicoded and u'content-type' in head.keys():
            content = head[u'content-type']
            if u'charset' not in content:
                head[u'content-type'] = content + u' ; charset=UTF-8'

        if head:
            headers = [str(headername) + ':' + str(headervalue)
                       for headername, headervalue in head.items()]
        else:
            headers = list()
        # Fix for expecting 100-continue from server, which not all servers
        # will send!
        headers.append("Expect:")
        headers.append("Connection: close")
        curl.setopt(curl.HTTPHEADER, headers)

        # reset the body, it holds values from previous runs otherwise
        headers = BytesIO()
        body = BytesIO()
        if sys.platform.find("win") > -1:
            curl.setopt(pycurl.CAINFO, libcurl_crt_file)
        curl.setopt(pycurl.WRITEFUNCTION, body.write)
        curl.setopt(pycurl.HEADERFUNCTION, headers.write)
        if verbose:
            curl.setopt(pycurl.VERBOSE, True)
        if ssl_insecure is True:
            curl.setopt(pycurl.SSL_VERIFYPEER, 0)
            curl.setopt(pycurl.SSL_VERIFYHOST, 0)

        curl.perform()  # Run the actual call

        response_body = body.getvalue()
        body.close()
        response_headers = headers.getvalue()
        headers.close()
        response_code = curl.getinfo(pycurl.RESPONSE_CODE)
        response = HttpResponse(
            body=response_body,
            headers=response_headers,
            status_code=response_code
        )
        self.response = response

        return response




#!/usr/bin/env python
import sys
import os
import json
import time
import logging
import threading
from notest.lib.utils import templated_var
from notest.http_test_exec import run_http_test
from notest.lib.utils import read_test_file
from notest.operations import get_operation_function
from notest.test_result import TestResult

ESCAPE_DECODING = 'unicode_escape'

sys.path.append(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__))))
from notest.context import Context
from notest.generators import parse_generator
from notest.operations import Operation
from notest.lib.parsing import flatten_dictionaries, lowercase_keys, \
    safe_to_bool, safe_to_json
from notest.http_test import HttpTest, DEFAULT_TIMEOUT


"""
Executable class, ties everything together into the framework.
Module responsibilities:
- Read & import test test_files
- Parse test configs
- Provide executor methods for sets of tests and benchmarks
- Collect and report on test/benchmark results
- Perform analysis on benchmark results
"""

logger = logging.getLogger('notest.master')


DIR_LOCK = threading.RLock()  # Guards operations changing the working directory


class CD:
    """Context manager for changing the current working directory"""

    # http://stackoverflow.com/questions/431684/how-do-i-cd-in-python/13197763#13197763

    def __init__(self, newPath):
        self.newPath = newPath

    def __enter__(self):
        if self.newPath:  # Don't CD to nothingness
            DIR_LOCK.acquire()
            self.savedPath = os.getcwd()
            os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        if self.newPath:  # Don't CD to nothingness            
            os.chdir(self.savedPath)
            DIR_LOCK.release()


class TestConfig:
    """ Configuration for a test run """
    timeout = DEFAULT_TIMEOUT  # timeout of tests, in seconds
    request_client = None  # requests or pycurl
    print_bodies = False  # Print response bodies in all cases
    print_headers = False  # Print response bodies in all cases
    retries = 0  # Retries on failures
    test_parallel = False  # Allow parallel execution of tests in a test set, for speed?
    interactive = False
    verbose = False
    ssl_insecure = False

    # Binding and creation of generators
    variable_binds = None
    generators = None  # Map of generator name to generator function

    def set_default_base_url(self, url):
        self.variable_binds['default_base_url'] = url

    def set_variable_binds(self, k, v):
        self.variable_binds[k] = v

    def __str__(self):
        return json.dumps(self, default=safe_to_json)


class TestSet:
    """ Encapsulates a set of tests and test configuration for them """
    tests = list()
    config = TestConfig()

    def __init__(self):
        self.config = TestConfig()
        self.tests = list()

    def __str__(self):
        return json.dumps(self, default=safe_to_json)


def parse_configuration(node, base_config=None):
    """ Parse input config to configuration information """
    test_config = base_config
    if not test_config:
        test_config = TestConfig()

    node = lowercase_keys(flatten_dictionaries(node))  # Make it usable

    # ahead process variable_binds, other key can use it in template
    if not test_config.variable_binds:
        test_config.variable_binds = dict()

    if 'variable_binds' in node:
        value = node['variable_binds']
        test_config.variable_binds.update(flatten_dictionaries(value))

    # set default_base_url to global variable_binds
    if 'default_base_url' in node:
        value = node['default_base_url']
        default_base_url = templated_var(value, test_config.variable_binds)
        test_config.variable_binds['default_base_url'] = default_base_url

    for key, value in node.items():
        if key == 'timeout':
            test_config.timeout = int(value)
        elif key == 'print_bodies':
            test_config.print_bodies = safe_to_bool(value)
        elif key == 'retries':
            test_config.retries = int(value)
        elif key == 'variable_binds':
            pass
        elif key == 'request_client':
            test_config.request_client = str(value)
        elif key == 'generators':
            flat = flatten_dictionaries(value)
            gen_map = dict()
            for generator_name, generator_config in flat.items():
                gen = parse_generator(generator_config,
                                      test_config.variable_binds)
                gen_map[str(generator_name)] = gen
            test_config.generators = gen_map

    return test_config


def parse_testsets(test_structure, test_files=set(), working_directory=None):
    """ Convert a Python data structure read from validated YAML to a set of structured testsets
    The data structure is assumed to be a list of dictionaries, each of which describes:
        - a tests (test structure)
        - a simple test (just a URL, and a minimal test is created)
        - or overall test configuration for this testset
        - an import (load another set of tests into this one, from a separate file)
            - For imports, these are recursive, and will use the parent config if none is present

    Note: test_files is used to track tests that import other tests, to avoid recursive loops

    This returns a list of testsets, corresponding to imported testsets and in-line multi-document sets
    """

    tests_list = list()
    test_config = TestConfig()
    testsets = list()

    if working_directory is None:
        working_directory = os.path.abspath(os.getcwd())

    # returns a testconfig and collection of tests
    for node in test_structure:  # Iterate through lists of test and configuration elements
        if isinstance(node, dict):  # Each config element is a miniature key-value dictionary
            node = lowercase_keys(node)
            for key in node:
                if key == 'import':
                    importfile = node[key]  # import another file
                    if importfile[0] != "/":
                        importfile = os.path.join(working_directory,
                                                  importfile)
                    if importfile not in test_files:
                        logger.info("Importing test sets: " + importfile)
                        test_files.add(importfile)
                        import_test_structure = read_test_file(importfile)
                        with CD(os.path.dirname(
                                os.path.realpath(importfile))):
                            import_testsets = parse_testsets(
                                import_test_structure, test_files)
                            testsets.extend(import_testsets)
                elif key == 'url':  # Simple test, just a GET to a URL
                    mytest = HttpTest()
                    val = node[key]
                    assert isinstance(val, str)
                    mytest.url = val
                    tests_list.append(mytest)
                elif key == 'test':  # Complex test with additional parameters
                    with CD(working_directory):
                        child = node[key]
                        mytest = HttpTest.parse_from_dict(child)
                        tests_list.append(mytest)
                elif key == 'operation':  # Complex test with additional parameters
                    operation = Operation()
                    operation.config = flatten_dictionaries(node[key])
                    tests_list.append(operation)
                elif key == 'config' or key == 'configuration':
                    test_config = parse_configuration(
                        node[key], base_config=test_config)

    testset = TestSet()
    testset.tests = tests_list
    testset.config = test_config
    for t in tests_list:
        t.testset_config = test_config
    testsets.append(testset)
    return testsets


def log_failure(failure, context=None, test_config=TestConfig()):
    """ Log a failure from a test """
    logger.error("Test Failure, failure type: {0}, Reason: {1}".format(
        failure.failure_type, failure.message))
    if failure.details:
        logger.error("Validator/Error details:" + str(failure.details))


def run_testsets(testsets):
    """ Execute a set of tests, using given TestSet list input """
    group_results = dict()  # results, by group
    group_failure_counts = dict()
    total_failures = 0
    myinteractive = False
    curl_handle = None

    for testset in testsets:
        mytests = testset.tests
        myconfig = testset.config
        context = Context()

        # Bind variables & add generators if pertinent
        if myconfig.variable_binds:
            context.bind_variables(myconfig.variable_binds)
        if myconfig.generators:
            for key, value in myconfig.generators.items():
                context.add_generator(key, value)

        # Make sure we actually have tests to execute
        if not mytests:
            # no tests in this test set, probably just imports.. skip to next
            # test set
            break

        myinteractive = True if myinteractive or myconfig.interactive else False

        # Run tests, collecting statistics as needed
        index = 0
        loop_count = 0

        while index < len(mytests) and loop_count < 100:
            test = mytests[index]
            if hasattr(test.testset_config, "loop_interval"):
                loop_interval = test.testset_config.loop_interval
            else:
                loop_interval = 2

            # Initialize the dictionaries to store test fail counts and results
            if test.group not in group_results:
                group_results[test.group] = list()
                group_failure_counts[test.group] = 0

            result = None
            if test.test_type == "http_test":
                result = run_http_test(test, test_config=myconfig, context=context,
                                       http_handler=curl_handle)
                result.body = None  # Remove the body, save some memory!

                if not result.passed:  # Print failure, increase failure counts for that test group
                    # Use result test URL to allow for templating
                    error_info = list()
                    error_info.append("")
                    error_info.append(' Test Failed: ' + test.name)
                    error_info.append(" URL=" + result.test.url)
                    error_info.append(" Group=" + test.group)
                    error_info.append(
                        " HTTP Status Code: " + str(result.response_code))
                    error_info.append("")
                    logger.error("\n".join(error_info))

                    # Print test failure reasons
                    if result.failures:
                        for failure in result.failures:
                            log_failure(failure, context=context,
                                        test_config=myconfig)

                    # Increment test failure counts for that group (adding an entry
                    # if not present)
                    failures = group_failure_counts[test.group]
                    failures = failures + 1
                    group_failure_counts[test.group] = failures

                else:  # Test passed, print results
                    msg = list()
                    msg.append("")
                    msg.append('Test Name: ' + test.name)
                    msg.append("  Request= {} {}".format(result.test.method, result.test.url))
                    msg.append("  Group=" + test.group)
                    msg.append("  HTTP Status Code: {}".format(result.response_code))
                    msg.append("  passed\n")
                    logger.info("\n".join(msg))

                # Add results for this test group to the resultset
                if not result.passed or result.loop is False:
                    group_results[test.group].append(result)

                # handle stop_on_failure flag
                if not result.passed and test.stop_on_failure is not None and test.stop_on_failure:
                    logger.info(
                        'STOP ON FAILURE! stopping test set execution, continuing with other test sets')
                    break
            elif test.test_type == "operation":
                logger.info("do operation {}, config:{}".format(
                    test.config.get('type'),
                    test.config
                ))
                result = TestResult()
                result.test_type = "operation"
                result.test = test
                try:
                    opt_name = test.config.get('type')
                    opt_func = get_operation_function(opt_name)
                    opt_func(test.config, context)
                    result.passed = True
                except Exception as e:
                    result.passed = False
                group_results[test.group].append(result)

            if result and result.loop is True:
                loop_count += 1
                time.sleep(loop_interval)
                continue
            else:
                index += 1
                continue

    if myinteractive:
        # a break for when interactive bits are complete, before summary data
        print("===================================")

    # Print summary results
    for group in sorted(group_results.keys()):
        test_count = len(group_results[group])
        failures = group_failure_counts[group]
        total_failures = total_failures + failures

        passfail = {True: 'SUCCEEDED: ', False: 'FAILED: '}
        output_string = "Test Group {0} {1}: {2}/{3} Tests Passed!".format(
            group, passfail[failures == 0], str(test_count - failures),
            str(test_count))

        print(output_string)

    return total_failures





from notest.common_test import CommonTest
OPERATIONS = dict()


def register_operations(typename, parse_function):
    """ Register a new tool for use in testing
        typename is the new tool type name (must not already exist)
        parse_function will run the tool
    """
    if not isinstance(typename, str):
        raise TypeError(
            'Generator type name {0} is invalid, must be a string'.format(typename))
    if typename in OPERATIONS:
        raise ValueError(
            'Generator type named {0} already exists'.format(typename))
    OPERATIONS[typename] = parse_function


def get_operation_function(typename):
    return OPERATIONS.get(typename.lower())


# Demo operation
def demo_operation(config, context=None):
    pass


class Operation(CommonTest):
    test_type = "operation"
    config = None
    group = "operation"



from __future__ import absolute_import
import sys
import string

"""
Parsing utilities, pulled out so they can be used in multiple modules
"""


def encode_unicode_bytes(my_string):
    """ Shim function, converts Unicode to UTF-8 encoded bytes regardless of the source format
        Intended for python 3 compatibility mode, and b/c PyCurl only takes raw bytes
    """
    if not isinstance(my_string, str):
        my_string = repr(my_string)

    if isinstance(my_string, str):
        return my_string.encode('utf-8')
    elif isinstance(my_string, bytes):
        return my_string


# TODO create a full class that extends string.Template
def safe_substitute_unicode_template(templated_string, variable_map):
    """ Perform string.Template safe_substitute on unicode input with unicode variable values by using escapes
        Catch: cannot accept unicode variable names, just values
        Returns a Unicode type output, if you want UTF-8 bytes, do encode_unicode_bytes on it
    """
    return string.Template(templated_string).safe_substitute(variable_map)


def safe_to_json(in_obj):
    """ Safely get dict from object if present for json dumping """
    if isinstance(in_obj, bytearray):
        return str(in_obj)
    if hasattr(in_obj, '__dict__'):
        return in_obj.__dict__
    try:
        return str(in_obj)
    except:
        return repr(in_obj)


def flatten_dictionaries(input):
    """ Flatten a list of dictionaries into a single dictionary, to allow flexible YAML use
      Dictionary comprehensions can do this, but would like to allow for pre-Python 2.7 use
      If input isn't a list, just return it.... """
    output = dict()
    if isinstance(input, list):
        for map in input:
            output.update(map)
    else:  # Not a list of dictionaries
        output = input
    return output


def lowercase_keys(input_dict):
    """ Take input and if a dictionary, return version with keys all lowercase and cast to str """
    if not isinstance(input_dict, dict):
        return input_dict
    safe = dict()
    for key, value in input_dict.items():
        safe[str(key).lower()] = value
    return safe


def safe_to_bool(input):
    """ Safely convert user input to a boolean, throwing exception if not boolean or boolean-appropriate string
      For flexibility, we allow case insensitive string matching to false/true values
      If it's not a boolean or string that matches 'false' or 'true' when ignoring case, throws an exception """
    if isinstance(input, bool):
        return input
    elif isinstance(input, str) and input.lower() == 'false':
        return False
    elif isinstance(input, str) and input.lower() == 'true':
        return True
    else:
        raise TypeError(
            'Input Object is not a boolean or string form of boolean!')




import sys
import logging
import json

logger = logging.Logger("mysql_generator")

from notest.lib.mysql_lib import MysqlClient
from notest.lib.utils import templated_var
from notest import generators


'''
 - generators:
        - task_id: {type: 'number_sequence', start: 10}
        - task_name:
            type: 'mysql'
            query: 'select name from sites'
            config: '$mysql_config'
'''


def parse_mysql_query_generator(config, variable_binds):
    """ Parses configuration options for a mysql_query generator """
    mysql_config = config.get('config')
    mysql_config = templated_var(mysql_config, variable_binds)
    if isinstance(mysql_config, str):
        mysql_config = json.loads(mysql_config)
    sql = config.get('query')
    sql = templated_var(sql)
    try:
        with MysqlClient(mysql_config) as cli:
            res = cli.query(sql)
            r = list()
            for i in res:
                if isinstance(i, tuple):
                    i = i[0]
                r.append(i)
            if len(r) == 0:
                raise Exception("No data queried in MySQL by '{}'!".format(sql))
            return generators.factory_fixed_sequence(r)()
    except Exception as e:
        logger.error(str(e))
        raise ValueError("Invalid query: " + sql + " : " + str(e))


GENERATORS = {'mysql': parse_mysql_query_generator}



import requests
import json
import os

from .http_auth_type import HttpAuthType
from .http_response import HttpResponse

base_dir = os.path.abspath(os.path.dirname(__file__))

DEFAULT_TIMEOUT = 10  # Seconds


class RequestsClient:
    def __init__(self, handler=None):
        if not handler or not isinstance(handler, requests.Session):
            self.session = requests.Session()
        else:
            self.session = handler
        self.response = None
        self.Http_Method_Map = {
            "post": self.session.post,
            "get": self.session.get,
            "put": self.session.put,
            "delete": self.session.delete,
            "multipart_post": self.session.post,
        }

    def get_handler(self):
        return self.session

    @staticmethod
    def close_handler(handler):
        if handler:
            handler.close()

    def close(self):
        if self.session:
            self.session.close()

    def send_request(self, test_obj, timeout=DEFAULT_TIMEOUT,
                     context=None,
                     handler=None, ssl_insecure=True, verbose=False):

        if handler:
            session = handler
        else:
            session = self.session

        # https://2.python-requests.org//zh_CN/latest/user/advanced.html
        request_obj = requests.Request(
            method=test_obj.method,
            url=str(test_obj.url),
        )
        request_obj = session.prepare_request(request_obj)

        if test_obj.auth_username and test_obj.auth_password:
            request_obj.auth = {test_obj.auth_username, test_obj.auth_password}

            if test_obj.auth_type:
                pass

        head = test_obj.get_headers(context=context)
        headers = {k.lower(): v for k, v in head.items()}
        # Set charset if doing unicode conversion and not set explicitly
        body = test_obj.http_body
        body_type = "data"
        if isinstance(body, str):  # Encode unicode
            # body = body.encode('UTF-8')
            if 'content-type' in headers.keys():
                content = headers['content-type']
                if 'json' in content:
                    body = json.loads(body)
                    body_type = "json"
                if 'charset' not in content:
                    headers['content-type'] = content + ' ; charset=UTF-8'

        if body_type == "json":
            request_obj.prepare_body(data=None, files=None, json=body)
        elif body_type == "data":
            request_obj.prepare_body(data=body, files=None, json=None)

        request_obj.headers.update(headers)


        verify = False
        if ssl_insecure is True:
            verify = False
        cert = None
        if "cert" in test_obj.__dict__ and test_obj.cert:
            cert = test_obj.cert
        proxies = None
        if "proxies" in test_obj.__dict__ and test_obj.proxies:
            proxies = test_obj.proxies

        if verbose:
            pass

        resp = session.send(
            request=request_obj,
            verify=verify,
            timeout=timeout,
            cert=cert,
            proxies=proxies
        )

        response = HttpResponse(
            requests_response=resp
        )
        self.response = response

        return response



import json
import sys

PYTHON_MAJOR_VERSION = sys.version_info[0]

import jmespath

try:  # First try to load notest from global namespace
    from notest import validators
    from notest import context
    from notest.lib import parsing
    from notest import contenthandling
except ImportError:  # Then try a relative import if possible
    from .. import validators
    from .. import context
    from .. import parsing
    from .. import contenthandling


class JMESPathExtractor(validators.AbstractExtractor):
    """ Extractor that uses JMESPath syntax
        See http://jmespath.org/specification.html for details
    """
    extractor_type = 'jmespath'
    is_body_extractor = True

    def extract_internal(self, body=None, headers=None, context=None):
        mybody = body
        if PYTHON_MAJOR_VERSION > 2:
            if isinstance(mybody, bytes):
                mybody = str(mybody, 'utf-8')

        try:
            res = jmespath.search(self.query, json.loads(mybody))  # Better way
            return res
        except Exception as e:
            raise ValueError("Invalid query: " + self.query + " : " + str(e))

    @classmethod
    def parse(cls, config):
        base = JMESPathExtractor()
        base.query = config
        return base


EXTRACTORS = {'jmespath': JMESPathExtractor.parse}



#!/usr/bin/env python
import sys
import os
import json
import logging
from optparse import OptionParser
from notest.lib.utils import read_test_file

sys.path.append(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__))))
from notest.lib.parsing import safe_to_bool
from notest.master import run_testsets, parse_testsets
from notest.plugin_registery import auto_load_ext
from notest.notest_lib import notest_run

"""
Executable class, ties everything together into the framework.
Module responsibilities:
- Read & import test test_files
- Parse test configs
- Provide executor methods for sets of tests and benchmarks
- Collect and report on test/benchmark results
- Perform analysis on benchmark results
"""
HEADER_ENCODING = 'ISO-8859-1'  # Per RFC 2616
LOGGING_LEVELS = {'debug': logging.DEBUG,
                  'info': logging.INFO,
                  'warning': logging.WARNING,
                  'error': logging.ERROR,
                  'critical': logging.CRITICAL}

DEFAULT_LOGGING_LEVEL = logging.INFO

logger = logging.getLogger('notest.main')
logging_config = {
    'level': DEFAULT_LOGGING_LEVEL,
    'format': "%(asctime)s - %(levelname)s - %(message)s"
}


def main(args):
    """
    Execute a test against the given base url.

    Keys allowed for args:
        test_file          - REQUIRED - Test file (yaml)
        log           - OPTIONAL - set logging level {debug,info,warning,error,critical} (default=warning)
        interactive   - OPTIONAL - mode that prints info before and after test exectuion and pauses for user input for each test
        config_file
        ssl_insecure
        ext_dir
        default_base_url
        request_client
        loop_interval

    """

    if 'log' in args and args['log'] is not None:
        level = LOGGING_LEVELS.get(args['log'].lower())
        logging_config['level'] = level
    logging.basicConfig(**logging_config)

    test_file = args['test_file']
    test_structure = read_test_file(test_file)

    args['test_structure'] = test_structure
    args['working_directory'] = os.path.dirname(test_file)

    # Execute all testsets
    failures_count = notest_run(args)

    sys.exit(failures_count)


def parse_command_line_args(args_in):
    """ Runs everything needed to execute from the command line, so main method is callable without arg parsing """
    parser = OptionParser(
        usage="usage: notest test_filen.yaml [options] ")
    parser.add_option("--log", help="Logging level",
                      action="store", type="string")
    parser.add_option("-i", "--interactive", help="Interactive mode",
                      action="store_true", dest="interactive", default=False)
    parser.add_option("-t", "--test-file", help="Test file to use, yaml or json file",
                      action="store", type="string",
                      dest="test_file")
    parser.add_option('--ssl-insecure',
                      help='Disable cURL host and peer cert verification',
                      action='store_true', default=False,
                      dest="ssl_insecure")
    parser.add_option('-e', '--ext-dir',
                      help='local extensions dir, default ./ext',
                      action='store',
                      dest="ext_dir")
    parser.add_option("-b", '--default-base-url',
                      help='default base url, if not specified, use the config in test file',
                      action='store',
                      dest="default_base_url")
    parser.add_option("-c", '--config-file',
                      help='config file',
                      action='store',
                      dest="config_file")
    parser.add_option("-l", '--loop-interval',
                      help='loop_interval, default 2s',
                      action='store',
                      dest="loop_interval")
    parser.add_option("-r", '--request-client',
                      help='request_client, select one in [requests, pycurl], default requests. '
                           'If use pycurl, you should install pycurl first by "pip install -U pycurl"',
                      action='store',
                      dest="request_client")
    parser.add_option("-v", '--override-config-variable-binds',
                      help='override_config_variable_binds, format -o key1=value1 -o key2=value2',
                      action='append',
                      dest="override_config_variable_binds")

    (args, unparsed_args) = parser.parse_args(args_in)
    args = vars(args)

    # Handle url/test as named, or, failing that, positional arguments
    if not args['test_file']:
        if len(unparsed_args) > 0:
            args['test_file'] = unparsed_args[0]
        else:
            parser.print_help()
            parser.error(
                "wrong number of arguments, need test filename, either as 1st parameters or via --test")

    # So modules can be loaded from current folder
    args['cwd'] = os.path.realpath(os.path.abspath(os.getcwd()))
    return args


def command_line_run():
    args_in = sys.argv[1:]
    args = parse_command_line_args(args_in)
    main(args)


# Allow import into another module without executing the main method
if __name__ == '__main__':
    command_line_run()



from pprint import pprint
import yaml

d = {
    "a": 1,
    "b": {
        "b1": 1,
        "b2": 2
    },
    "c": [
        1, 2, 3
    ],
    "d": [
        {
            "d1": 1,
            "d2": 2
         }
    ]
}

y = """
- test:
    - generator_binds:
        - task_id: task_id
        - name: name
    - group: "Quickstart"
    - name: "post"
    - url: "/tasks"
    - method: "POST"
    - headers: {'Content-Type': 'application/json', "Token": 123}
    - body: {template: '{"title": "$name", "id": $task_id, "done": $done}'}
    - expected_status: [201]
    - validators:
        # Test key does not exist
        - extract_test: {jsonpath_mini: "info.0.title", test: "exists"}
        - compare:
            - jsonpath_mini: "info.0.id"
            - comparator: "str_eq"
            - expected: {template: '$task_id'}
    - extract_binds:
        - post_task_id: {jsonpath_mini: 'info.0.id'}
        - post_task_title:
            mysql:
              sql: 'select name from sites limit 1'
              config: '{"user": "root", "password": "password", "host": "192.168.99.101", "database": "test"}'
"""

# print(yaml.dump(d))
pprint(yaml.load(y))



from notest.lib.utils import templated_var
import time

'''
- operation:
    - type: "sleep_operation"
    - seconds: 5
'''


def sleep_operation(config, context=None):
    print("Run sleep_operation")
    assert isinstance(config, dict)
    assert "seconds" in config
    s = config['seconds']
    s = templated_var(s, context)
    print("Sleep {} s".format(s))
    time.sleep(int(s))
    print()


OPERATIONS = {'sleep_operation': sleep_operation}



#!/usr/bin/env python
import sys
import os
import traceback
import logging
from notest.clients.request_client import get_client_class
from notest.http_test import HttpTestResult, parse_headers, HttpTest

ESCAPE_DECODING = 'unicode_escape'

sys.path.append(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__))))
from notest.context import Context
from notest import validators

from notest.validators import Failure

logger = logging.getLogger('notest.http_test')


HEADER_ENCODING = 'ISO-8859-1'  # Per RFC 2616


def run_http_test(mytest, test_config, context=None,
                  http_handler=None, *args, **kwargs):
    """ Put together test pieces: configure & run actual test, return results """
    assert isinstance(mytest, HttpTest)

    # Initialize a context if not supplied
    my_context = context
    if my_context is None:
        my_context = Context()
    mytest.context = my_context
    mytest.update_context_before()

    result = HttpTestResult()
    result.test = mytest
    result.passed = None

    if test_config.interactive:
        print("===================================")
        print("%s" % mytest.name)
        print("-----------------------------------")
        print("REQUEST:")
        print("%s %s" % (mytest.method, mytest.url))
        print("HEADERS:")
        print("%s" % (mytest.headers))
        if mytest.body is not None:
            print("\n%s" % mytest.body)

        input("Press ENTER when ready")

    # send request
    try:
        http_response = mytest.send_request(
            timeout=test_config.timeout,
            context=my_context,
            handler=http_handler,
            ssl_insecure=test_config.ssl_insecure,
            verbose=test_config.verbose
        )
    except Exception as e:
        # Curl exception occurred (network error), do not pass go, do not
        # collect $200
        trace = traceback.format_exc()
        result.failures.append(Failure(message="Curl Exception: {0}".format(
            e), details=trace,
            failure_type=validators.FAILURE_CURL_EXCEPTION))
        result.passed = False
        client = mytest.testset_config.request_client
        if not client:
            client = "requests"
        HttpClient = get_client_class(client)
        HttpClient.close_handler(http_handler)
        return result

    # Retrieve Body
    result.body = http_response.body

    # Retrieve Headers
    headers = http_response.headers
    if headers and isinstance(headers, bytes):
        headers = str(headers, HEADER_ENCODING)  # Per RFC 2616
        # Parse HTTP headers
        try:
            result.response_headers = parse_headers(headers)
        except Exception as e:
            trace = traceback.format_exc()
            error = "Header parsing exception: {} {}".format(e, headers)
            result.failures.append(
                Failure(
                    message=error,
                    details=trace,
                    failure_type=validators.FAILURE_TEST_EXCEPTION))
            result.passed = False
            return result
    elif headers and isinstance(headers, list):
        pass
    elif headers and isinstance(headers, dict):
        pass
    else:
        error = "Unknown Header Type: {}".format(type(headers))
        error_detail = "Unknown Header Type: {} {}".format(type(headers), headers)
        result.failures.append(
            Failure(
                message=error,
                details=error_detail,
                failure_type=validators.FAILURE_TEST_EXCEPTION))
        result.passed = False
        return result
    result.response_headers = headers
    response_code = http_response.status_code
    result.response_code = response_code

    logger.debug("Initial Test Result, based on expected response code: " +
                 str(response_code in mytest.expected_status))

    if response_code in mytest.expected_status:
        result.passed = True
    else:
        # Invalid response code
        result.passed = False
        failure_message = "Invalid HTTP response code: response code {0} not in expected codes [{1}]".format(
            response_code, mytest.expected_status)
        result.failures.append(Failure(
            message=failure_message, details=None,
            failure_type=validators.FAILURE_INVALID_RESPONSE))

    # print str(test_config.print_bodies) + ',' + str(not result.passed) + ' ,
    # ' + str(test_config.print_bodies or not result.passed)

    headers = result.response_headers

    # execute validator
    if result.passed is True:
        body = result.body
        if mytest.validators is not None and isinstance(mytest.validators,
                                                        list):
            logger.debug("executing validators: " +
                         str(len(mytest.validators)))
            failures = result.failures
            for validator in mytest.validators:
                validate_result = validator.validate(
                    body=body, headers=headers, context=my_context)
                if not validate_result:
                    result.passed = False
                # Proxy for checking if it is a Failure object, because of
                # import issues with isinstance there
                if isinstance(validate_result, validators.Failure):
                    failures.append(validate_result)
                # TODO add printing of validation for interactive mode
        else:
            logger.debug("no validators found")

        # Only do context updates if test was successful
        mytest.update_context_after(result.body, headers)

    # execute loop_until_conditions
    if result.passed is True:
        body = result.body
        if mytest.loop_until_conditions is not None and isinstance(mytest.loop_until_conditions, list):
            logger.debug("executing loop_until_conditions: " +
                         str(len(mytest.loop_until_conditions)))
            result.loop = False
            for validator in mytest.loop_until_conditions:
                validate_result = validator.validate(
                    body=body, headers=headers, context=my_context)
                if isinstance(validate_result, validators.Failure):
                    result.loop = True
                    logger.error(validate_result)
        else:
            logger.debug("no loop_until_conditions found")

    # Print response body if override is set to print all *OR* if test failed
    # (to capture maybe a stack trace)
    if test_config.print_bodies or not result.passed:
        if test_config.interactive:
            print("RESPONSE:")
        if result.body:
            body = result.body
            if isinstance(body, bytes):
                print(body.decode())
            else:
                print(body)
        # else:
        #     print("None")

    if test_config.print_headers or not result.passed:
        if test_config.interactive:
            print("RESPONSE HEADERS:")
        if result.response_headers:
            print(result.response_headers)
        # else:
        #     print("None")

    # TODO add string escape on body output
    logger.debug(str(result))

    return result



import mysql.connector
from mysql.connector import errorcode
import logging

logger = logging.Logger("mysql_lib")


class MysqlClient:
    def __init__(self, config):
        self.connection = None
        self.cursor = None
        if not config:
            raise Exception("No MySQL config")
        self.config = config

    def __enter__(self):
        self.connect(self.config)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self, config=None):
        if not config:
            config = self.config
        try:
            connection = mysql.connector.connect(**config)
            self.connection = connection
            self.cursor = self.get_cursor()
        except mysql.connector.Error as err:
            if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                logger.error(
                    "Something is wrong with your user name or password")
            elif err.errno == errorcode.ER_BAD_DB_ERROR:
                logger.error("Database does not exist")
            elif err.errno == errorcode.CR_CONN_HOST_ERROR:
                logger.error("Host Connection Error")
            else:
                logger.error(err.errno)
            raise err
        return self.connection

    def close(self):
        try:
            self.connection.close()
            logger.debug("Connection Closed")
        except Exception as e:
            logger.error(str(e))

    def get_cursor(self):
        if not self.cursor:
            if not self.connection:
                self.connection = self.connect(None)
            self.cursor = self.connection.cursor()
        return self.cursor

    def execute(self, sql):
        try:
            cursor = self.get_cursor()
            cursor.execute(sql)
            self.connection.commit()
        except Exception as e:
            logger.error(str(e))
            raise e

    def query(self, sql):
        cursor = self.get_cursor()
        cursor.execute(sql)
        ret = cursor.fetchall()
        return ret


if __name__ == '__main__':
    # docker run -it --name mysql -p 3306:3306 -p 33060:33060 -e MYSQL_ROOT_PASSWORD=password -e MYSQLOST=0.0.0.0 mysql:5.7
    config = {
        'user': 'root',
        'password': 'password',
        'host': '192.168.99.101',
        'database': 'test',
        'use_pure': False
    }

    with MysqlClient(config) as cli:
        # cli.execute(
        #     "insert into sites(name, url) values('baidu', 'www.baidu.com')")
        # cli.execute(
        #     "insert into sites(name, url) values('google', 'www.google.com')")
        ret = cli.query("select * from sites where name='baidu'")
        print(type(ret))
        for i in ret:
            print(i)



import random
import string
import os
import logging

from notest.lib.parsing import flatten_dictionaries, lowercase_keys

""" Collection of generators to be used in templating for test data

Plans: extend these by allowing generators that take generators for input
Example: generators that case-swap
"""

INT32_MAX_VALUE = 2147483647  # Max of 32 bit unsigned int

logger = logging.getLogger('notest.generators')

# Character sets to use in text generation, python string plus extras
CHARACTER_SETS = {
    'ascii_letters': string.ascii_letters,
    'ascii_lowercase': string.ascii_lowercase,
    'ascii_uppercase': string.ascii_uppercase,
    'digits': string.digits,
    'hexdigits': string.hexdigits,
    'hex_lower':  string.digits + 'abcdef',
    'hex_upper':  string.digits + 'ABCDEF',
    'letters': string.ascii_letters,
    'lowercase': string.ascii_lowercase,
    'octdigits': string.octdigits,
    'punctuation': string.punctuation,
    'printable': string.printable,
    'uppercase': string.ascii_uppercase,
    'whitespace': string.whitespace,
    'url.slug': string.ascii_lowercase + string.digits + '-',
    'url.safe': string.ascii_letters + string.digits + '-~_.',
    'alphanumeric': string.ascii_letters + string.digits,
    'alphanumeric_lower': string.ascii_lowercase + string.digits,
    'alphanumeric_upper': string.ascii_uppercase + string.digits
}


def factory_generate_ids(starting_id=1, increment=1):
    """ Return function generator for ids starting at starting_id
        Note: needs to be called with () to make generator """
    def generate_started_ids():
        val = starting_id
        local_increment = increment
        while(True):
            yield val
            val += local_increment
    return generate_started_ids


def generator_basic_ids():
    """ Return ids generator starting at 1 """
    return factory_generate_ids(1)()


def generator_random_int32():
    """ Random integer generator for up to 32-bit signed ints """
    rand = random.Random()
    while (True):
        yield random.randint(0, INT32_MAX_VALUE)


def factory_generate_text(legal_characters=string.ascii_letters, min_length=8, max_length=8):
    """ Returns a generator function for text with given legal_characters string and length
        Default is ascii letters, length 8

        For hex digits, combine with string.hexstring, etc
        """
    def generate_text():
        local_min_len = min_length
        local_max_len = max_length
        rand = random.Random()
        while(True):
            length = random.randint(local_min_len, local_max_len)
            array = [random.choice(legal_characters)
                     for x in range(0, length)]
            yield ''.join(array)

    return generate_text


def factory_fixed_sequence(values):
    """ Return a generator that runs through a list of values in order, looping after end """

    def seq_generator():
        my_list = list(values)
        i = 0
        while(True):
            yield my_list[i]
            i += 1
            if i == len(my_list):
                i = 0
    return seq_generator


def parse_fixed_sequence(config):
    """ Parse fixed sequence string """
    vals = config['values']
    if not vals:
        raise ValueError('Values for fixed sequence must exist')
    if not isinstance(vals, list):
        raise ValueError('Values must be a list of entries')
    return factory_fixed_sequence(vals)()


def factory_choice_generator(values):
    """ Return a generator that picks values from a list randomly """

    def choice_generator():
        my_list = list(values)
        rand = random.Random()
        while(True):
            yield random.choice(my_list)
    return choice_generator


def parse_choice_generator(config):
    """ Parse choice generator """
    vals = config['values']
    if not vals:
        raise ValueError('Values for choice sequence must exist')
    if not isinstance(vals, list):
        raise ValueError('Values must be a list of entries')
    return factory_choice_generator(vals)()


def factory_env_variable(env_variable):
    """ Return a generator function that reads from an environment variable """

    def return_variable():
        variable_name = env_variable
        while(True):
            yield os.environ.get(variable_name)

    return return_variable


def factory_env_string(env_string):
    """ Return a generator function that uses OS expand path to expand environment variables in string """

    def return_variable():
        my_input = env_string
        while(True):
            yield os.path.expandvars(my_input)

    return return_variable

""" Implements the parsing logic for YAML, and acts as single point for reading configuration """


def parse_random_text_generator(configuration):
    """ Parses configuration options for a random text generator """
    character_set = configuration.get(u'character_set')
    characters = None
    if character_set:
        character_set = character_set.lower()
        if character_set not in CHARACTER_SETS:
            raise ValueError(
                "Illegal character set name, is not defined: {0}".format(character_set))
        characters = CHARACTER_SETS[character_set]
    else:  # Custom characters listing, not a character set
        characters = str(configuration.get(u'characters'))

    min_length = 8
    max_length = 8

    if configuration.get(u'min_length'):
        min_length = int(configuration.get(u'min_length'))
    if configuration.get(u'max_length'):
        max_length = int(configuration.get(u'max_length'))

    if configuration.get(u'length'):
        length = int(configuration.get(u'length'))
        min_length = length
        max_length = length

    if characters:
        return factory_generate_text(legal_characters=characters, min_length=min_length, max_length=max_length)()
    else:
        return factory_generate_text(min_length=min_length, max_length=max_length)()


# List of valid generator types
GENERATOR_TYPES = set(['env_variable',
                       'env_string',
                       'number_sequence',
                       'random_int',
                       'random_text',
                       'fixed_sequence'
                       ])

GENERATOR_PARSING = {'fixed_sequence': parse_fixed_sequence}


def register_generator(typename, parse_function):
    """ Register a new generator for use in testing
        typename is the new generator type name (must not already exist)
        parse_function will parse a configuration object (dict)
    """
    if not isinstance(typename, str):
        raise TypeError(
            'Generator type name {0} is invalid, must be a string'.format(typename))
    if typename in GENERATOR_TYPES:
        raise ValueError(
            'Generator type named {0} already exists'.format(typename))
    GENERATOR_TYPES.add(typename)
    GENERATOR_PARSING[typename] = parse_function

# Try registering a new generator
register_generator('choice', parse_choice_generator)


def parse_generator(configuration, variable_binds=None):
    """ Parses a configuration built from yaml and returns a generator
        Configuration should be a map
    """

    configuration = lowercase_keys(flatten_dictionaries(configuration))
    gen_type = str(configuration.get('type')).lower()

    if gen_type not in GENERATOR_TYPES:
        raise ValueError(
            'Generator type given {0} is not valid '.format(gen_type))

    # Do the easy parsing, delegate more complex logic to parsing functions
    if gen_type == 'env_variable':
        return factory_env_variable(configuration['variable_name'])()
    elif gen_type == 'env_string':
        return factory_env_string(configuration['string'])()
    elif gen_type == 'number_sequence':
        start = configuration.get('start')
        increment = configuration.get('increment')
        if not start:
            start = 1
        else:
            start = int(start)
        if not increment:
            increment = 1
        else:
            increment = int(increment)
        return factory_generate_ids(start, increment)()
    elif gen_type == 'random_int':
        return generator_random_int32()
    elif gen_type == 'random_text':
        return parse_random_text_generator(configuration)
    elif gen_type in GENERATOR_TYPES:
        return GENERATOR_PARSING[gen_type](configuration, variable_binds)
    else:
        raise Exception("Unknown generator type: {0}".format('gen_type'))



from flask import Flask, jsonify, request
import json
import time

app = Flask(__name__)

tasks = [
    {
        'id': 1,
        'title': 'aaaaa',
        'done': False
    }
]
token = "123"


def get_max_id(extend_items=None):
    t_id = -1
    for i in tasks:
        if t_id < i['id']:
            t_id = i['id']
    if extend_items:
        for i in extend_items:
            if t_id < i['id']:
                t_id = i['id']
    return t_id


@app.route('/token', methods=['POST'])
def post_token():
    global token
    token = str(time.time())
    res = {"status": "Created",
           "Token": token}
    print(res)
    return json.dumps(res), 201


@app.route('/clear_all', methods=['POST'])
def clear_all():
    print("Parameters: ", request.args)
    print("Headers: ", request.headers)
    global token
    token = str(time.time())
    tasks.clear()
    res = {"status": "Cleared"}
    print(res)
    return json.dumps(res), 204


@app.route('/get_form_response', methods=['GET'])
def get_form_response():
    return "param1=1&param2=2", 200


@app.route('/tasks_form', methods=['POST'])
def tasks_form_operations():
    print("Parameters: ", request.args)
    print("Headers: ", request.headers)
    global tasks, token

    token_t = request.headers.get('Token', None)
    # if token_t != token:
    if token_t == None:
        res = {"status": "ERROR",
               "msg": "Permission Denied, No Token specified",
               "info": request.args}
        return jsonify(res), 401

    if request.method == "POST":
        body_s = request.get_data()
        if isinstance(body_s, bytes):
            body_s = body_s.decode()
        form = request.form
        print("form: ", type(form), form)
        task_id = form.get("id", None)
        if task_id == None or task_id == "":
            task_id = get_max_id() + 1
        tmp = None
        task_id = int(task_id)
        for t in tasks:
            if t["id"] == task_id:
                tmp = t
                break
        if tmp != None:
            res = {"status": "ERROR",
                   "msg": "ID %s existed" % task_id,
                   "info": body_s}
            return jsonify(res), 400
        title = form.get("title")
        done = form.get("done")
        done = True if done == 'True' else False
        new_task = {"id": task_id, "title": title, "done": done}
        tasks.append(new_task)
        res = {"status": "Created",
               "info": new_task}
        return jsonify(res), 201


@app.route('/tasks', methods=['GET', 'POST'])
def tasks_operations():
    print("Parameters: ", request.args)
    print("Headers: ", request.headers)
    global tasks, token

    token_t = request.headers.get('Token', None)
    # if token_t != token:
    if token_t == None:
        res = {"status": "ERROR",
               "msg": "Permission Denied, No Token specified",
               "info": request.args}
        return jsonify(res), 401

    if request.method == "GET":
        return jsonify({'info': tasks})

    elif request.method == "POST":
        body_s = request.get_data()
        if isinstance(body_s, bytes):
            body_s = body_s.decode()
        print("body: ", body_s)
        try:
            body = json.loads(body_s)
        except:
            res = {"status": "ERROR",
                   "msg": "body is not json type: %s" % body_s,
                   "info": body_s}
            return jsonify(res), 400
        if type(body) != list:
            body = [body]
        items_added = list()
        for body_item in body:
            task_id = body_item.get("id", None)
            if task_id == None or task_id == "":
                task_id = get_max_id(items_added) + 1
            tmp = None
            task_id = int(task_id)
            for t in tasks:
                if t["id"] == task_id:
                    tmp = t
                    break
            if tmp != None:
                res = {"status": "ERROR",
                       "msg": "ID %s existed" % task_id,
                       "info": body_item}
                return jsonify(res), 400
            title = body_item.get("title")
            done = body_item.get("done")
            new_task = {"id": task_id, "title": title, "done": done}
            items_added.append(new_task)
        tasks.extend(items_added)
        res = {"status": "Created",
               "info": items_added}
        return jsonify(res), 201


@app.route('/task/<int:task_id>', methods=['GET', 'PUT', 'DELETE'])
def task_operations(task_id):
    print("Parameters: ", request.args)
    print("Headers: ", request.headers)
    print("Body: ", request.data)
    global tasks, token

    token_t = request.headers.get('Token', None)
    # if token_t != token:
    if token_t == None:
        res = {"status": "ERROR",
               "msg": "Permission Denied, No Token specified",
               "info": request.args}
        return jsonify(res), 401

    if request.method == "GET":
        task = [t for t in tasks if t['id'] == task_id]
        if len(task) == 0:
            res = {"status": "ERROR",
                   "msg": "ID %s not exist" % task_id,
                   "info": request.url}
            return jsonify(res), 404
        return jsonify({'info': task})

    elif request.method == "POST":
        body_s = request.get_data()
        print("body: ", body_s)
        try:
            body = json.loads(body_s)
        except:
            res = {"status": "ERROR",
                   "msg": "body is not json type: %s" % body_s,
                   "info": body_s}
            return jsonify(res), 400
        tmp = None
        for t in tasks:
            if t["id"] == task_id:
                tmp = t
                break
        if tmp != None:
            res = {"status": "ERROR",
                   "msg": "ID %s existed" % task_id,
                   "info": body_s}
            return jsonify(res), 400
        title = body.get("title", None)
        done = body.get("done", None)
        new_task = {"id": task_id, "title": title, "done": done}
        tasks.append(new_task)
        res = {"status": "Created",
               "info": new_task}
        return jsonify(res), 201

    elif request.method == "PUT":
        body_s = request.get_data()
        if isinstance(body_s, bytes):
            body_s = body_s.decode()
        print("body: ", body_s)
        try:
            body = json.loads(body_s)
        except:
            res = {"status": "ERROR",
                   "msg": "body is not json type",
                   "info": body_s}
            return jsonify(res), 400
        task = None
        for t in tasks:
            if t["id"] == task_id:
                task = t
                break
        if task == None:
            res = {"status": "ERROR",
                   "msg": "ID %s not exist" % task_id,
                   "info": body_s}
            return jsonify(res), 404
        for key, value in list(body.items()):
            if key != "id":
                task[key] = value
        res = {"status": "Updated",
               "info": task}
        return jsonify(res), 201

    elif request.method == "DELETE":
        task_index = None
        for i, t in enumerate(tasks):
            if t["id"] == task_id:
                task_index = i
                break
        if task_index == None:
            res = {"status": "ERROR",
                   "msg": "ID %s not exist" % task_id,
                   "info": request.args}
            return jsonify(res), 400
        ret = tasks[task_index].copy()
        del tasks[task_index]
        res = {"status": "Deleted",
               "info": ret}
        return jsonify(res), 200


@app.route('/echo_request_body', methods=['POST'])
def echo_request_body():
    body = request.get_data()
    if isinstance(body, bytes):
        body = body.decode()
    return body, 200


ready_time = None
@app.route('/delay_task', methods=['GET', 'POST'])
def delay_task():
    global ready_time
    print("Parameters: ", request.args)
    print("Headers: ", request.headers)
    print("Body: ", request.data)

    if request.method == "GET":
        if time.time() > ready_time:
            resp = {"state": "ready"}
            print(resp)
            return jsonify(resp), 200
        else:
            resp = {"state": "running"}
            print(resp)
            return jsonify(resp), 200
    elif request.method == "POST":
        ready_time = time.time() + 5
        resp = {"state": "ok"}
        print(resp)
        return jsonify(resp), 201


if __name__ == '__main__':
    # app.run(debug=True, port=5000, host='0.0.0.0', ssl_context='adhoc')
    app.run(debug=True, port=5000, host='0.0.0.0')


import json
import logging
import string
from email import message_from_string

logger = logging.getLogger('notest.http_test')

from notest.clients.request_client import get_client_class
from notest.clients.http_auth_type import HttpAuthType
from notest.lib.utils import templated_var
from notest.test_result import TestResult
from notest.common_test import CommonTest
from notest.lib.parsing import lowercase_keys, flatten_dictionaries, safe_to_bool
import notest.validators as validators

"""
Pull out the Test objects and logic associated with them
This module implements the internal responsibilities of a test object:
- Test parameter/configuration storage
- Templating for tests
- Parsing of test configuration from results of YAML read
"""

DEFAULT_TIMEOUT = 10  # Seconds


# Parsing helper functions
def coerce_to_string(val):
    if isinstance(val, str):
        return val
    elif isinstance(val, int):
        return str(val)
    elif isinstance(val, bytes):
        return val.decode('utf-8')
    else:
        raise TypeError(
            "Input {0} is not a string or integer, and it needs to be!".format(
                val))


def coerce_string_to_ascii(val):
    if isinstance(val, str):
        return val.encode('ascii')
    elif isinstance(val, bytes):
        return val
    else:
        raise TypeError(
            "Input {0} is not a string, string expected".format(val))


def coerce_http_method(val):
    myval = val
    if not isinstance(myval, str) or len(val) == 0:
        raise TypeError(
            "Invalid HTTP method name: input {0} is not a string or has 0 length".format(
                val))
    if isinstance(myval, bytes):
        myval = myval.decode('utf-8')
    return myval.upper()


def coerce_list_of_ints(val):
    """ If single value, try to parse as integer, else try to parse as list of integer """
    if isinstance(val, list):
        return [int(x) for x in val]
    else:
        return [int(val)]


class HttpTestResult(TestResult):
    """ Encapsulates everything about a test response """
    test = None  # Test run
    response_code = None

    body = None  # Response body, if tracked

    passed = False
    response_headers = None
    failures = None
    loop = False

    def __init__(self):
        self.failures = list()

    def __str__(self):
        msg = list()
        msg.append("\n====================")
        msg.append("Test Type: {}".format(self.test.test_type))
        msg.append("Passed? : {}".format(self.passed))
        msg.append("Test Url: {} {}".format(self.test.method, self.test.url))
        msg.append("Response Code: {}".format(self.response_code))
        msg.append("Response Headers: {}".format(self.response_headers))
        msg.append("Response Body: {}".format(self.body))
        msg.append("Failures : {}".format(self.failures))
        msg.append("====================\n")

        return "\n".join(msg)


def parse_headers(header_string):
    """ Parse a header-string into individual headers
        Implementation based on: http://stackoverflow.com/a/5955949/95122
        Note that headers are a list of (key, value) since duplicate headers are allowed

        NEW NOTE: keys & values are unicode strings, but can only contain ISO-8859-1 characters
    """
    # First line is request line, strip it out
    if not header_string:
        return list()
    request, headers = header_string.split('\r\n', 1)
    if not headers:
        return list()

    header_msg = message_from_string(headers)
    # Note: HTTP headers are *case-insensitive* per RFC 2616
    return [(k.lower(), v) for k, v in header_msg.items()]


class HttpTest(CommonTest):
    """ Describes a REST test """
    test_type = "http_test"
    url = None
    expected_status = [200]  # expected HTTP status code or codes
    http_body = None
    http_headers = dict()  # HTTP Headers
    method = 'GET'
    group = 'Default'
    name = 'Unnamed'
    validators = None  # Validators for response body, IE regexes, etc
    loop_until_conditions = None
    stop_on_failure = True
    failures = None
    auth_username = None
    auth_password = None
    auth_type = HttpAuthType.HTTP_AUTH_BASIC
    delay = 0

    templates = None  # Dictionary of template to compiled template

    # Bind variables, generators, and contexts
    variable_binds = None
    generator_binds = None  # Dict of variable name and then generator name
    extract_binds = None  # Dict of variable name and extract function to run
    context = None

    def __init__(self):
        self.headers = dict()
        self.expected_status = [200]
        self.http_client = None
        self.http_handler = None

    @staticmethod
    def has_contains():
        return 'contains' in validators.VALIDATORS

    def ninja_copy(self):
        """ Optimization: limited copy of test object, for realize() methods
            This only copies fields changed vs. class, and keeps methods the same
        """
        output = HttpTest()
        myvars = vars(self)
        output.__dict__ = myvars.copy()
        return output

    # These are variables that can be templated
    def set_body(self, value):
        """ Set body, directly """
        self.http_body = value

    def get_body(self, context=None):
        """ Read body from file, applying template if pertinent """
        if self.http_body is None:
            return None
        elif isinstance(self.http_body, str):
            return self.http_body
        else:
            return self.http_body.get_content(context=context)

    body = property(get_body, set_body, None,
                    'Request body, if any (for POST/PUT methods)')

    def set_context(self, context=None):
        self.context = context

    NAME_HEADERS = 'headers'

    def set_headers(self, value):
        self.http_headers = value

    def get_headers(self, context=None):
        """ Get headers, applying template if pertinent """
        if not context:
            context = self.context
        if not context:
            return self.http_headers

        vals = context.get_values()

        def template_tuple(tuple_input):
            return (string.Template(str(tuple_item)).safe_substitute(vals)
                    for tuple_item in tuple_input)

        return dict(map(template_tuple, self.http_headers.items()))

    headers = property(get_headers, set_headers, None,
                       'Headers dictionary for request')

    def update_context_before(self):
        """ Make pre-test context updates, by applying variable and generator updates """
        context = self.context
        if self.variable_binds:
            context.bind_variables(self.variable_binds)
        if self.generator_binds:
            for key, value in self.generator_binds.items():
                context.bind_generator_next(key, value)

    def update_context_after(self, response_body, headers):
        """ Run the extraction routines to update variables based on HTTP response body """
        context = self.context
        if self.extract_binds:
            for key, value in self.extract_binds.items():
                result = value.extract(
                    body=response_body, headers=headers, context=context)
                context.bind_variable(key, result)

    def realize(self, context=None):
        if not context:
            context = self.context
        if self.url.startswith('/'):
            self.url = "$default_base_url" + self.url
        self.url = templated_var(self.url, context)
        self.method = templated_var(self.method, context)
        self.body = templated_var(self.body, context)
        self.headers = templated_var(self.headers, context)

    def send_request(self, timeout=DEFAULT_TIMEOUT, context=None,
                     handler=None, ssl_insecure=True, verbose=False):
        if not context:
            context = self.context
        self.realize(context)

        client = self.testset_config.request_client
        if not client:
            client = "requests"
        HttpClient = get_client_class(client)
        self.http_client = HttpClient(handler)
        self.http_handler = self.http_client.get_handler()
        return self.http_client.send_request(
            test_obj=self,
            timeout=timeout,
            context=context,
            ssl_insecure=ssl_insecure,
            verbose=verbose
        )

    @classmethod
    def parse_from_dict(cls, node, input_test=None):
        """ Create or modify a test, input_test, using configuration in node, and base_url
        If no input_test is given, creates a new one

        Test_path gives path to test file, used for setting working directory in setting up input bodies

        Uses explicitly specified elements from the test input structure
        to make life *extra* fun, we need to handle list <-- > dict transformations.

        This is to say: list(dict(),dict()) or dict(key,value) -->  dict() for some elements

        Accepted structure must be a single dictionary of key-value pairs for test configuration """

        mytest = input_test
        if not mytest:
            mytest = HttpTest()

        # Clean up for easy parsing
        node = lowercase_keys(flatten_dictionaries(node))

        # Simple table of variable name, coerce function, and optionally special store function
        CONFIG_ELEMENTS = {
            # Simple variables
            u'auth_username': [coerce_string_to_ascii],
            u'auth_password': [coerce_string_to_ascii],
            u'method': [coerce_http_method],  # HTTP METHOD
            u'delay': [lambda x: int(x)],  # Delay before running
            u'group': [coerce_to_string],  # Test group name
            u'name': [coerce_to_string],  # Test name
            u'expected_status': [coerce_list_of_ints],
            u'stop_on_failure': [safe_to_bool],

            # Templated / special handling
            # u'body': [ContentHandler.parse_content]

            # COMPLEX PARSE OPTIONS
            # u'extract_binds':[],  # Context variable-to-extractor output binding
            # u'variable_binds': [],  # Context variable to value binding
            # u'generator_binds': [],  # Context variable to generator output binding
            # u'validators': [],  # Validation functions to run
        }

        def use_config_parser(configobject, configelement, configvalue):
            """ Try to use parser bindings to find an option for parsing and storing config element
                :configobject: Object to store configuration
                :configelement: Configuratione element name
                :configvalue: Value to use to set configuration
                :returns: True if found match for config element, False if didn't
            """

            myparsing = CONFIG_ELEMENTS.get(configelement)
            if myparsing:
                converted = myparsing[0](configvalue)
                setattr(configobject, configelement, converted)
                return True
            return False

        # Copy/convert input elements into appropriate form for a test object
        for configelement, configvalue in node.items():
            if use_config_parser(mytest, configelement, configvalue):
                continue

            # Configure test using configuration elements
            if configelement == 'url':
                if isinstance(configvalue, dict):
                    configvalue = configvalue.get("template")
                mytest.url = coerce_to_string(configvalue)
            elif configelement == 'body':
                if isinstance(configvalue, dict):
                    configvalue = configvalue.get("template")
                mytest.http_body = configvalue
            elif configelement == 'extract_binds':
                # Add a list of extractors, of format:
                # {variable_name: {extractor_type: extractor_config}, ... }
                binds = flatten_dictionaries(configvalue)
                if mytest.extract_binds is None:
                    mytest.extract_binds = dict()

                for variable_name, extractor in binds.items():
                    if not isinstance(extractor, dict) or len(
                            extractor) == 0:
                        raise TypeError(
                            "Extractors must be defined as maps of extractorType:{configs} with 1 entry")
                    if len(extractor) > 1:
                        raise ValueError(
                            "Cannot define multiple extractors for given variable name")

                    # Safe because length can only be 1
                    for extractor_type, extractor_config in extractor.items():
                        mytest.extract_binds[
                            variable_name] = validators.parse_extractor(
                            extractor_type, extractor_config)

            elif configelement == 'validators':
                # Add a list of validators
                if not isinstance(configvalue, list):
                    raise Exception(
                        'Misconfigured validator section, must be a list of validators')
                if mytest.validators is None:
                    mytest.validators = list()

                # create validator and add to list of validators
                for var in configvalue:
                    if not isinstance(var, dict):
                        raise TypeError(
                            "Validators must be defined as validatorType:{configs} ")
                    for validator_type, validator_config in var.items():
                        validator = validators.parse_validator(
                            validator_type, validator_config)
                        mytest.validators.append(validator)

            elif configelement == 'loop_until':
                # Add a list of validators
                if not isinstance(configvalue, list):
                    raise Exception(
                        'Misconfigured validator section, must be a list of validators')
                if mytest.loop_until_conditions is None:
                    mytest.loop_until_conditions = list()

                # create validator and add to list of validators
                for var in configvalue:
                    if not isinstance(var, dict):
                        raise TypeError(
                            "Validators must be defined as validatorType:{configs} ")
                    for validator_type, validator_config in var.items():
                        validator = validators.parse_validator(
                            validator_type, validator_config)
                        mytest.loop_until_conditions.append(validator)

            elif configelement == 'headers':  # HTTP headers to use, flattened to a single string-string dictionary
                configvalue = flatten_dictionaries(configvalue)
                if isinstance(configvalue, dict):
                    mytest.headers = configvalue
                elif isinstance(configvalue, str):
                    try:
                        mytest.headers = json.loads(configvalue)
                    except Exception as e:
                        logger.error(str(e))
                        raise Exception("header must be dict or json str")
                else:
                    raise TypeError(
                        "Illegal header type: headers must be a dictionary or list of dictionary keys")
            elif configelement == 'variable_binds':
                mytest.variable_binds = flatten_dictionaries(configvalue)
            elif configelement == 'generator_binds':
                output = flatten_dictionaries(configvalue)
                output2 = dict()
                for key, value in output.items():
                    output2[str(key)] = str(value)
                mytest.generator_binds = output2

        # For non-GET requests, accept additional response codes indicating success
        # (but only if not expected statuses are not explicitly specified)
        # this is per HTTP spec:
        # http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.5
        if 'expected_status' not in node.keys():
            if mytest.method == 'POST':
                mytest.expected_status = [200, 201, 204]
            elif mytest.method == 'PUT':
                mytest.expected_status = [200, 201, 204]
            elif mytest.method == 'DELETE':
                mytest.expected_status = [200, 202, 204]
            # Fallthrough default is simply [200]
        return mytest



from notest.notest_lib import notest_run

args = {
    'config_file': '../examples\\config.json',
    # 'default_base_url': None,
    'override_config_variable_binds': {
        'title': 'GodQ'
    },
    # 'ext_dir': None,
    'loop_interval': 1,
    # 'request_client': None,
    # 'working_directory': '../examples',
    'test_structure': [{'config': {'default_base_url': 'http://localhost:5000',
                                   'generators': [{'id': {'start': 10,
                                                          'type': 'number_sequence'}}],
                                   'testset': 'Quickstart app tests',
                                   'variable_binds': {'done': 'true',
                                                      'title': 'Gaius-Test'}}},
                       {'test': {'expected_status': [201],
                                 'method': 'POST',
                                 'name': 'post ready task',
                                 'url': '/delay_task',
                                 'body': '$title'}},
                       {'test': {'expected_status': 200,
                                 'headers': {'Content-Type': 'application/json',
                                             'Token': 123},
                                 'loop_until': [{'extract_test': {'jsonpath_mini': 'state',
                                                                  'test': 'exists'}},
                                                {'compare': {'comparator': 'str_eq',
                                                             'expected': 'ready',
                                                             'jsonpath_mini': 'state'}}],
                                 'method': 'GET',
                                 'name': 'get ready task',
                                 'url': '/delay_task',
                                 'body': '{"title": "$title"}'}}]

}

fc = notest_run(args)


import json
from notest.lib.mysql_lib import MysqlClient
from notest.lib.utils import templated_var
from notest import validators


'''
- extract_binds:
        - post_task_id: {jsonpath_mini: 'info.0.id'}
        - post_task_title:
            mysql:
              query: 'select name from sites limit 1'
              config: '$mysql_config'
'''


class MySQLQueryExtractor(validators.AbstractExtractor):
    """
    Extractor that uses MySQL SQL query syntax
    """
    extractor_type = 'mysql'

    def __init__(self):
        self.query = None
        self.mysql_config = None

    def extract_internal(self, body=None, headers=None, context=None):
        self.query = templated_var(self.query, context)
        self.mysql_config = templated_var(self.mysql_config, context)
        self.mysql_config = json.loads(self.mysql_config)
        try:
            with MysqlClient(self.mysql_config) as cli:
                res = cli.query(self.query)
                if len(res) == 0:
                    raise Exception(
                        "No data queried in MySQL by '{}'!".format(self.query))
                res = res[0]
                if isinstance(res, tuple):
                    res = res[0]
                return res
        except Exception as e:
            raise ValueError("Invalid query: '" + self.query + "' : " + str(e))

    @classmethod
    def parse(cls, config):
        mysql_config = config.get('config')
        sql = config.get('query')
        entity = MySQLQueryExtractor()
        entity.mysql_config = mysql_config
        entity.query = sql
        return entity


EXTRACTORS = {'mysql': MySQLQueryExtractor.parse}



import sys
import os
import json
import logging
from notest.lib.utils import read_test_file

sys.path.append(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__))))
from notest.lib.parsing import safe_to_bool
from notest.master import run_testsets, parse_testsets
from notest.plugin_registery import auto_load_ext

"""
Executable class, ties everything together into the framework.
Module responsibilities:
- Read & import test test_files
- Parse test configs
- Provide executor methods for sets of tests and benchmarks
- Collect and report on test/benchmark results
- Perform analysis on benchmark results
"""
HEADER_ENCODING = 'ISO-8859-1'  # Per RFC 2616
LOGGING_LEVELS = {'debug': logging.DEBUG,
                  'info': logging.INFO,
                  'warning': logging.WARNING,
                  'error': logging.ERROR,
                  'critical': logging.CRITICAL}

DEFAULT_LOGGING_LEVEL = logging.INFO

logger = logging.getLogger('notest.main')
logging_config = {
    'level': DEFAULT_LOGGING_LEVEL,
    'format': "%(asctime)s - %(message)s"
}

CONFIG = None


def load_config(config_file):
    global CONFIG
    if not CONFIG:
        if os.path.isfile(config_file):
            with open(config_file, "r") as fd:
                data = fd.read()
                if isinstance(data, bytes):
                    data = data.decode()
                data = json.loads(data)
                CONFIG = data
    return CONFIG


def notest_run(args):
    """
        Execute a test against the given base url.

        Keys allowed for args:
            test_structure          - REQUIRED - Test file (yaml/json)
            working_directory      - OPTIONAL
            override_config_variable_binds  - OPTIONAL - override variable_binds of config in test file
            interactive   - OPTIONAL - mode that prints info before and after test exectuion and pauses for user input for each test
                                     please set False when not used by command
            config_file   - OPTIONAL
            ssl_insecure   - OPTIONAL  default True
            ext_dir   - OPTIONAL
            default_base_url   - OPTIONAL
            request_client   - OPTIONAL  default requests
            loop_interval   - OPTIONAL   default 2s
        """
    # import pprint
    # pprint.pprint(args)

    test_structure = args.get("test_structure")
    assert test_structure

    config_file = None
    if 'config_file' in args and args['config_file'] is not None:
        config_file = args['config_file']
    else:
        config_file = "config.json"
    config_from_file = load_config(config_file)
    if config_from_file:
        for k, v in args.items():
            if v:
                config_from_file[k] = v
        args = config_from_file

    working_directory = None
    if 'working_directory' in args and args['working_directory']:
        working_directory = args['working_directory']

    testsets = parse_testsets(test_structure,
                              working_directory=working_directory)

    # Override configs from command line if config set
    for testset in testsets:
        if 'interactive' in args and args['interactive'] is not None:
            ia = args['interactive']
            if isinstance(ia, str):
                ia = safe_to_bool(args['interactive'])
            assert isinstance(ia, bool)
            testset.config.interactive = ia

        if 'verbose' in args and args['verbose'] is not None:
            testset.config.verbose = safe_to_bool(args['verbose'])

        if 'ssl_insecure' in args and args['ssl_insecure'] is not None:
            testset.config.ssl_insecure = safe_to_bool(args['ssl_insecure'])

        if 'ext_dir' in args and args['ext_dir'] is not None:
            if not os.path.exists(args['ext_dir']):
                msg = "Plugin Folder ext_dir can not found, path:'{}'  ...... Skipped\n".format(args['ext_dir'])
                logger.error(msg)
            elif os.path.isdir(args['ext_dir']):
                auto_load_ext(args['ext_dir'])
            else:
                msg = "Option ext_dir must be folder, path:'{}'".format(args['ext_dir'])
                logger.error(msg)
                raise Exception(msg)

        if 'default_base_url' in args and args['default_base_url'] is not None:
            testset.config.set_default_base_url(args['default_base_url'])

        if 'override_config_variable_binds' in args and args['override_config_variable_binds'] is not None:
            override_vars = args['override_config_variable_binds']
            if isinstance(override_vars, bytes):
                override_vars = override_vars.decode()
            if isinstance(override_vars, str):
                try:
                    override_vars = json.loads(override_vars)
                except Exception as e:
                    logger.error("Load json error, {} {} ".format(str(e), override_vars))
                    raise e
            elif isinstance(override_vars, list):
                vd = dict()
                for v in override_vars:
                    t = v.split("=")
                    if len(t) != 2:
                        raise Exception("Option override_config_variable_binds can not be parsed! {}".format(override_vars))
                    vd[t[0]] = t[1]
                override_vars = vd
            assert isinstance(override_vars, dict)
            for k, v in override_vars.items():
                testset.config.set_variable_binds(k, v)

        if 'request_client' in args and args['request_client'] is not None and not testset.config.request_client:
            testset.config.request_client = args['request_client']

        if 'loop_interval' in args and args['loop_interval']:
            testset.config.loop_interval = int(args['loop_interval'])

    # Execute all testsets
    failures_count = run_testsets(testsets)

    return failures_count


import logging
import types

"""
Basic context implementation for binding variables to values
"""

logger = logging.getLogger('notest.context')


class Context(object):
    """ Manages binding of variables & generators, with both variable name and generator name being strings """

    variables = dict()  # Maps variable name to current value
    generators = dict()  # Maps generator name to generator function
    mod_count = 0  # Lets us see if something has been altered, avoiding needless retemplating

    def bind_variable(self, variable_name, variable_value):
        """ Bind a named variable to a value within the context
            This allows for passing in variables in testing """
        str_name = str(variable_name)
        prev = self.variables.get(str_name)
        if prev != variable_value:
            self.variables[str(variable_name)] = variable_value
            self.mod_count = self.mod_count + 1
            #logging.debug('Context: altered variable named {0} to value {1}'.format(str_name, variable_value))

    def bind_variables(self, variable_map):
        for key, value in variable_map.items():
            self.bind_variable(key, value)

    def add_generator(self, generator_name, generator):
        """ Adds a generator to the context, this can be used to set values for a variable
            Once created, you can set values with the generator via bind_generator_next """

        if not isinstance(generator, types.GeneratorType):
            raise ValueError(
                'Cannot add generator named {0}, it is not a generator type'.format(generator_name))

        self.generators[str(generator_name)] = generator
        #logging.debug('Context: Added generator named {0}'.format(generator_name))

    def bind_generator_next(self, variable_name, generator_name):
        """ Binds the next value for generator_name to variable_name and return value used """
        str_gen_name = str(generator_name)
        str_name = str(variable_name)
        val = next(self.generators[str_gen_name])

        prev = self.variables.get(str_name)
        if prev != val:
            self.variables[str_name] = val
            self.mod_count = self.mod_count + 1
            # Logging is /expensive/
            #logging.debug('Context: Set variable named {0} to next value {1} from generator named {2}'.format(variable_name, val, generator_name))
        return val

    def get_values(self):
        return self.variables

    def get_value(self, variable_name):
        """ Get bound variable value, or return none if not set """
        return self.variables.get(str(variable_name))

    def get_generators(self):
        return self.generators

    def get_generator(self, generator_name):
        return self.generators.get(str(generator_name))

    def __init__(self):
        self.variables = dict()
        self.generators = dict()



import json

from notest.lib.utils import templated_var
from notest.lib.mysql_lib import MysqlClient

'''
- operation:
    - type: "mysql_upsert"
    - config: '{"user": "root", "password": "password", "host": "192.168.99.101", "database": "test"}'
    - sql: 'insert into sites(name, url) values("a", "a.com")'
'''


def mysql_upsert(config, context=None):
    print("Run mysql_upsert")
    assert isinstance(config, dict)
    assert "config" in config
    assert "sql" in config
    sql = templated_var(config['sql'], context)
    mysql_config = templated_var(config['config'], context)
    mysql_config = json.loads(mysql_config)
    with MysqlClient(mysql_config) as cli:
        cli.execute(sql)


OPERATIONS = {'mysql_upsert': mysql_upsert}



import logging
import json
import operator
import traceback
import os
import re

# Local module imports
from notest.lib import parsing
from notest.lib.utils import templated_var

"""
Validator/Extractor logic for utility use
Defines objects:
- Extractors that take a text body and context, and return a result from the (text) body
- Validators (functions) that take a text body and context, and validate the body
- Several useful implementations for each:

Extractors:
    - json mini extractor (sudo-jsonpath)

Validators:
    - TEST validator, config includes and extractor function and test name, applies test to extract results
        - Uses VALIDATOR_TESTS, for pluggable test functions
    - comparator validator:
        - runs named extractor, compares to expected value (can be template or extractor)
        - uses (pluggable) comparator function for comparison

"""

logger = logging.getLogger('notest.validators')

# Binary comparison tests
COMPARATORS = {
    'count_eq': lambda x, y: safe_length(x) == y,
    'lt': operator.lt,
    'less_than': operator.lt,
    'le': operator.lt,
    'less_than_or_equal': operator.lt,
    'eq': operator.eq,
    'equals': operator.eq,
    'str_eq': lambda x, y: operator.eq(str(x), str(y)),
    'ne': operator.ne,
    'not_equals': operator.ne,
    'ge': operator.ge,
    'greater_than_or_equal': operator.ge,
    'gt': operator.gt,
    'greater_than': operator.gt,
    'contains': lambda x, y: x and operator.contains(x, y),  # is y in x
    'contained_by': lambda x, y: y and operator.contains(y, x),  # is x in y
    'regex': lambda x, y: regex_compare(str(x), str(y)),
    'type': lambda x, y: test_type(x, y)
}
COMPARATORS['length_eq'] = COMPARATORS['count_eq']

# Allow for testing basic types in comparators
TYPES = {
    'null': type(None),
    'none': type(None),
    'number': (int, float),
    'int': int,
    'float': float,
    'boolean': bool,
    'string': str,
    'array': list,
    'list': list,
    'dict': dict,
    'map': dict,
    'scalar': (bool, int, float, str, type(None)),
    'collection': (list, dict, set)
}


def test_type(val, mytype):
    """ Check value fits one of the types, if so return true, else false """

    typelist = TYPES.get(mytype.lower())

    if typelist is None:
        raise TypeError(
            "Type {0} is not a valid type to test against!".format(
                mytype.lower()))
    try:
        for testtype in typelist:
            if isinstance(val, testtype):
                return True
        return False
    except TypeError:
        return isinstance(val, typelist)


# Unury comparison tests
VALIDATOR_TESTS = {
    'exists': lambda x: x is not None,
    'not_exists': lambda x: x is None
}

# Validators and Extractors are registered once their parse functions exist
EXTRACTORS = dict()
VALIDATORS = dict()


def safe_length(var):
    """ Exception-safe length check, returns -1 if no length on type or error """
    output = -1
    try:
        output = len(var)
    except:
        pass
    return output


def regex_compare(input, regex):
    return bool(re.search(regex, input))


# Validator Failure Reasons
FAILURE_INVALID_RESPONSE = 'Invalid HTTP Response Code'
FAILURE_CURL_EXCEPTION = 'Curl Exception'
FAILURE_TEST_EXCEPTION = 'Test Execution Exception'
FAILURE_VALIDATOR_FAILED = 'Validator Failed'
FAILURE_VALIDATOR_EXCEPTION = 'Validator Exception'
FAILURE_EXTRACTOR_EXCEPTION = 'Extractor Exception'


class Failure(object):
    """ Encapsulates why and how a validation failed for user consumption
        Message is a short explanation, details is a longer, multiline reason
        Validator is the validator that failed (for config info)
    """
    message = None
    failure_type = None
    details = None
    validator = None

    def __nonzero__(self):
        """ Failure objects test as False, simplifies coding with them """
        return False

    def __bool__(self):
        """ Failure objects test as False, simplifies coding with them """
        return False

    def __str__(self):
        return self.message

    def __init__(self, message="", details="", failure_type=None,
                 validator=None):
        self.message = message
        self.details = details
        self.validator = validator
        self.failure_type = failure_type


class AbstractExtractor(object):
    """ Basic extractor, you only need to implement full_extract """

    extractor_type = None
    query = None
    is_body_extractor = False  # Uses response body
    is_header_extractor = False  # Uses response headers

    def __str__(self):
        return "Extractor type: {0}, query: {1} ".format(
            self.extractor_type, self.query)

    def extract_internal(self, body=None, headers=None, context=None):
        """ Do extraction, query should be pre-templated """
        pass

    def extract(self, body=None, headers=None, context=None):
        """ Extract data """
        self.query = templated_var(data=self.query, context=context)
        return self.extract_internal(body=body, headers=headers, context=context)

    def get_readable_config(self, context=None):
        """ Print a human-readable version of the configuration """
        query = templated_var(data=self.query, context=context)
        output = 'Extractor Type: {0},  Query: "{1}"'.format(
            self.extractor_type, query)
        return output


class MiniJsonExtractor(AbstractExtractor):
    """ Extractor that uses jsonpath_mini syntax
        IE key.key or array_index.key extraction
    """
    extractor_type = 'jsonpath_mini'
    is_body_extractor = True

    def extract_internal(self, body=None, headers=None, context=None):
        if isinstance(body, bytes):
            body = body.decode('utf-8')  # Default JSON encoding

        try:
            body = json.loads(body)
            return self.query_dictionary(self.query, body)
        except ValueError:
            raise ValueError("Not legal JSON!")

    @staticmethod
    def query_dictionary(query, dictionary, delimiter='.'):
        """ Do an xpath-like query with dictionary, using a template if relevant """
        # Based on
        # http://stackoverflow.com/questions/7320319/xpath-like-query-for-nested-python-dictionaries

        try:
            stripped_query = query.strip(delimiter)
            if stripped_query:
                for x in stripped_query.split(delimiter):
                    try:
                        x = int(x)
                        dictionary = dictionary[x]
                    except ValueError:
                        dictionary = dictionary[x]
        except:
            return None
        return dictionary

    @classmethod
    def parse(cls, config):
        base = MiniJsonExtractor()
        base.query = config
        return base


class HeaderExtractor(AbstractExtractor):
    """ Extractor that pulls out a named header value... or list of values if multiple values defined """
    extractor_type = 'header'
    is_header_extractor = True

    def extract_internal(self, body=None, headers=None, context=None):
        low = query.lower()
        # Value for all matching key names
        extracted = [y[1] for y in filter(lambda x: x[0] == low, headers)]
        if len(extracted) == 0:
            raise ValueError("Invalid header name {0}".format(self.query))
        elif len(extracted) == 1:
            return extracted[0]
        else:
            return extracted

    @classmethod
    def parse(cls, config, extractor_base=None):
        base = HeaderExtractor()
        base.query = config
        return base


class RawBodyExtractor(AbstractExtractor):
    """ Extractor that returns the full request body """
    extractor_type = 'raw_body'
    is_header_extractor = False
    is_body_extractor = True

    def extract_internal(self, query=None, args=None, body=None,
                         headers=None, context=None):
        return body

    @classmethod
    def parse(cls, config, extractor_base=None):
        # Doesn't take any real configuration
        base = RawBodyExtractor()
        return base


def _get_extractor(config_dict):
    """ Utility function, get an extract function for a single valid extractor name in config
        and error if more than one or none """
    extractor = None
    extract_config = None
    for key, value in config_dict.items():
        if key in EXTRACTORS:
            return parse_extractor(key, value)
    else:  # No valid extractor
        raise Exception(
            'No valid extractor name to use in input: {0}'.format(
                config_dict))


class AbstractValidator(object):
    """ Encapsulates basic validator handling """
    name = None
    config = None

    def validate(self, body=None, headers=None, context=None):
        """ Run the validation function, return true or a Failure """
        pass


class ComparatorValidator(AbstractValidator):
    """ Does extract and compare from request body   """

    name = 'ComparatorValidator'
    config = None  # Configuration text, if parsed
    extractor = None
    comparator = None
    comparator_name = ""
    expected = None
    isTemplateExpected = False

    def get_readable_config(self, context=None):
        """ Get a human-readable config string """
        string_frags = list()
        string_frags.append(
            "Extractor: " + self.extractor.get_readable_config(
                context=context))
        if isinstance(self.expected, AbstractExtractor):
            string_frags.append("Expected value extractor: " +
                                self.expected.get_readable_config(
                                    context=context))
        elif self.isTemplateExpected:
            string_frags.append(
                'Expected is templated, raw value: {0}'.format(
                    self.expected))
        return os.linesep.join(string_frags)

    def validate(self, body=None, headers=None, context=None):
        try:
            extracted_val = self.extractor.extract(
                body=body, headers=headers, context=context)
        except Exception as e:
            trace = traceback.format_exc()
            return Failure(message="Extractor threw exception",
                           details=trace, validator=self,
                           failure_type=FAILURE_EXTRACTOR_EXCEPTION)

        # Compute expected output, either templating or using expected value
        expected_val = None
        if isinstance(self.expected, AbstractExtractor):
            try:
                expected_val = self.expected.extract(body=body, headers=headers, context=context)
            except Exception as e:
                trace = traceback.format_exc()
                return Failure(
                    message="Expected value extractor threw exception",
                    details=trace, validator=self,
                    failure_type=FAILURE_EXTRACTOR_EXCEPTION)
        elif context:
            expected_val = templated_var(self.expected, context)
        else:
            expected_val = self.expected

        # Handle a bytes-based body and a unicode expected value seamlessly
        if isinstance(extracted_val, bytes) and isinstance(
                expected_val, str):
            expected_val = expected_val.encode('utf-8')
        comparison = self.comparator(extracted_val, expected_val)

        if not comparison:
            failure = Failure(validator=self)
            failure.message = "Comparison failed, evaluating {0}({1}, {2}) returned False".format(
                self.comparator_name, extracted_val, expected_val)
            failure.details = self.get_readable_config(context=context)
            failure.failure_type = FAILURE_VALIDATOR_FAILED
            return failure
        else:
            return True

    @staticmethod
    def parse(config):
        """ Create a validator that does an extract from body and applies a comparator,
            Then does comparison vs expected value
            Syntax sample:
              { jsonpath_mini: 'node.child',
                operator: 'eq',
                expected: 'myValue'
              }
        """

        output = ComparatorValidator()
        config = parsing.lowercase_keys(
            parsing.flatten_dictionaries(config))
        output.config = config

        # Extract functions are called by using defined extractor names
        output.extractor = _get_extractor(config)

        if output.extractor is None:
            raise ValueError(
                "Extract function for comparison is not valid or not found!")

        if 'comparator' not in config:  # Equals comparator if unspecified
            output.comparator_name = 'eq'
        else:
            output.comparator_name = config['comparator'].lower()
        output.comparator = COMPARATORS[output.comparator_name]
        if not output.comparator:
            raise ValueError("Invalid comparator given!")

        try:
            expected = config['expected']
        except KeyError:
            raise ValueError(
                "No expected value found in comparator validator config, one must be!")

        # Expected value can be another extractor query, or a single value, or
        # a templated value

        if isinstance(expected, str) or isinstance(expected, (int, float, complex)):
            output.expected = expected
        elif isinstance(expected, dict):
            expected = parsing.lowercase_keys(expected)
            template = expected.get('template')
            if template:  # Templated string
                if not isinstance(template, str):
                    raise ValueError(
                        "Can't template a comparator-validator unless template value is a string")
                output.expected = template
            else:  # Extractor to compare against
                output.expected = _get_extractor(expected)
                if not output.expected:
                    raise ValueError(
                        "Can't supply a non-template, non-extract dictionary to comparator-validator")

        return output


class ExtractTestValidator(AbstractValidator):
    """ Does extract and test from request body """
    name = 'ExtractTestValidator'
    extractor = None
    test_fn = None
    test_name = None
    config = None

    def get_readable_config(self, context=None):
        """ Get a human-readable config string """
        return "Extractor: " + self.extractor.get_readable_config(
            context=context)

    @staticmethod
    def parse(config):
        output = ExtractTestValidator()
        config = parsing.lowercase_keys(
            parsing.flatten_dictionaries(config))
        output.config = config
        extractor = _get_extractor(config)
        output.extractor = extractor

        test_name = config['test']
        output.test_name = test_name
        test_fn = VALIDATOR_TESTS[test_name]
        output.test_fn = test_fn
        return output

    def validate(self, body=None, headers=None, context=None):
        try:
            extracted = self.extractor.extract(
                body=body, headers=headers, context=context)
        except Exception as e:
            trace = traceback.format_exc()
            return Failure(
                message="Exception thrown while running extraction from body",
                details=trace, validator=self,
                failure_type=FAILURE_EXTRACTOR_EXCEPTION)

        tested = self.test_fn(extracted)
        if tested:
            return True
        else:
            failure = Failure(details=self.get_readable_config(
                context=context), validator=self,
                failure_type=FAILURE_VALIDATOR_FAILED)
            failure.message = "Extract and test validator failed on test: {0}({1})".format(
                self.test_name, extracted)
            # TODO can we do better with details?
            return failure


def parse_extractor(extractor_type, config):
    """ Convert extractor type and config to an extractor instance
        Uses registered parse function for that extractor type
        Parse functions may return either:
            - An extraction function (wrapped in an Extractor instance with configs and returned)
            - OR a a full Extractor instance (configured)
    """
    parse = EXTRACTORS.get(extractor_type.lower())
    if not parse:
        raise ValueError(
            "Extractor {0} is not a valid extractor type".format(
                extractor_type))
    parsed = parse(config)

    if isinstance(parsed,
                  AbstractExtractor):  # Parser gave a full extractor
        return parsed

    # Look for matching attributes... simple inheritance has issues because of
    # cross-module loading
    items = AbstractExtractor().__dict__
    if set(parsed.__dict__.keys()).issuperset(set(items.keys())):
        return parsed
    else:
        raise TypeError(
            "Parsing functions for extractors must return an AbstractExtractor instance!")


def parse_validator(name, config_node):
    '''Parse a validator from configuration and use it '''
    name = name.lower()
    if name not in VALIDATORS:
        raise ValueError(
            "Name {0} is not a named validator type!".format(name))
    valid = VALIDATORS[name](config_node)

    if valid.name is None:  # Carry over validator name if none set in parser
        valid.name = name
    if valid.config is None:  # Store config info if absent
        valid.config = config_node
    return valid


def register_validator(name, parse_function):
    ''' Registers a validator for use by this library
        Name is the string name for validator

        Parse function does parse(config_node) and returns a Validator object
        Validator functions have signature:
            validate(response_body, context=None) - context is a bindings.Context object

        Validators return true or false and optionally can return a Failure instead of false
        This allows for passing more details
    '''
    name = name.lower()
    if name in VALIDATORS:
        raise Exception("Validator exists with this name: {0}".format(name))

    VALIDATORS[name] = parse_function


def register_extractor(extractor_name, parse_function):
    """ Register a new body extraction function """
    if not isinstance(extractor_name, str):
        raise TypeError("Cannot register a non-string extractor name")
    if extractor_name.lower() == 'comparator':
        raise ValueError(
            "Cannot register extractors called 'comparator', that is a reserved name")
    elif extractor_name.lower() == 'test':
        raise ValueError(
            "Cannot register extractors called 'test', that is a reserved name")
    elif extractor_name.lower() == 'expected':
        raise ValueError(
            "Cannot register extractors called 'expected', that is a reserved name")
    elif extractor_name in EXTRACTORS:
        raise ValueError(
            "Cannot register an extractor name that already exists: {0}".format(
                extractor_name))
    EXTRACTORS[extractor_name] = parse_function


def register_test(test_name, test_function):
    """ Register a new one-argument test function """
    if not isinstance(test_name, str):
        raise TypeError("Cannot register a non-string test name")
    elif test_name in VALIDATOR_TESTS:
        raise ValueError(
            "Cannot register a test name that already exists: {0}".format(
                test_name))
    VALIDATOR_TESTS[test_name] = test_function


def register_comparator(comparator_name, comparator_function):
    """ Register a new twpo-argument comparator function returning true or false """
    if not isinstance(comparator_name, str):
        raise TypeError("Cannot register a non-string comparator name")
    elif comparator_name in COMPARATORS:
        raise ValueError(
            "Cannot register a comparator name that already exists: {0}".format(
                comparator_name))
    COMPARATORS[comparator_name] = comparator_function


# --- REGISTRY OF EXTRACTORS AND VALIDATORS ---
register_extractor('jsonpath_mini', MiniJsonExtractor.parse)
register_extractor('header', HeaderExtractor.parse)
register_extractor('raw_body', RawBodyExtractor.parse)
# ENHANCEME: add JsonPath-rw support for full JsonPath syntax
# ENHANCEME: add elementree support for xpath extract on XML, very simple no?
# See: https://docs.python.org/2/library/xml.etree.elementtree.html,
# findall syntax

register_validator('comparator', ComparatorValidator.parse)
register_validator('compare', ComparatorValidator.parse)
register_validator('assertEqual', ComparatorValidator.parse)
register_validator('extract_test', ExtractTestValidator.parse)
register_validator('assertTrue', ExtractTestValidator.parse)



import os

from notest.lib.parsing import *

# Python 2/3 switches
PYTHON_MAJOR_VERSION = sys.version_info[0]
if PYTHON_MAJOR_VERSION > 2:
    pass


"""
Encapsulates contend handling logic, for pulling file content into tests
"""


class ContentHandler:
    """ Handles content that may be (lazily) read from filesystem and/or templated to various degrees
    Also creates pixie dust and unicorn farts on demand
    This is pulled out because logic gets complex rather fast

    Covers 6 states:
        - Inline body content, no templating
        - Inline body content, with templating
        - File path to content, NO templating
        - File path to content, content gets templated
        - Templated path to file content (path itself is templated), file content UNtemplated
        - Templated path to file content (path itself is templated), file content TEMPLATED
    """

    content = None  # Inline content
    is_file = False
    is_template_path = False
    is_template_content = False

    def is_dynamic(self):
        """ Is templating used? """
        return self.is_template_path or self.is_template_content

    def get_content(self, context=None):
        """ Does all context binding and pathing to get content, templated out """

        if self.is_file:
            path = self.content
            if self.is_template_path and context:
                path = string.Template(path).safe_substitute(
                    context.get_values())
            data = None
            with open(path, 'r') as f:
                data = f.read()

            if self.is_template_content and context:
                return string.Template(data).safe_substitute(context.get_values())
            else:
                return data
        else:
            if self.is_template_content and context:
                return safe_substitute_unicode_template(self.content, context.get_values())
            else:
                return self.content

    def create_noread_version(self):
        """ Read file content if it is static and return content handler with no I/O """
        if not self.is_file or self.is_template_path:
            return self
        output = ContentHandler()
        output.is_template_content = self.is_template_content
        with open(self.content, 'r') as f:
            output.content = f.read()
        return output

    def setup(self, input, is_file=False, is_template_path=False, is_template_content=False):
        """ Self explanatory, input is inline content or file path. """
        if not isinstance(input, str):
            raise TypeError("Input is not a string")
        if is_file:
            input = os.path.abspath(input)
        self.content = input
        self.is_file = is_file
        self.is_template_path = is_template_path
        self.is_template_content = is_template_content

    @staticmethod
    def parse_content(node):
        """ Parse content from input node and returns ContentHandler object
        it'll look like:

            - template:
                - file:
                    - temple: path

            or something

        """

        # Tread carefully, this one is a bit narly because of nesting
        output = ContentHandler()
        is_template_path = False
        is_template_content = False
        is_file = False
        is_done = False

        while (node and not is_done):  # Dive through the configuration tree
            # Finally we've found the value!
            if isinstance(node, str):
                output.content = node
                output.setup(node, is_file=is_file, is_template_path=is_template_path,
                             is_template_content=is_template_content)
                return output
            elif not isinstance(node, dict) and not isinstance(node, list):
                raise TypeError(
                    "Content must be a string, dictionary, or list of dictionaries")

            is_done = True

            # Dictionary or list of dictionaries
            flat = lowercase_keys(flatten_dictionaries(node))
            for key, value in flat.items():
                if key == u'template':
                    if isinstance(value, str):
                        if is_file:
                            value = os.path.abspath(value)
                        output.content = value
                        is_template_content = is_template_content or not is_file
                        output.is_template_content = is_template_content
                        output.is_template_path = is_file
                        output.is_file = is_file
                        return output
                    else:
                        is_template_content = True
                        node = value
                        is_done = False
                        break

                elif key == 'file':
                    if isinstance(value, str):
                        output.content = os.path.abspath(value)
                        output.is_file = True
                        output.is_template_content = is_template_content
                        return output
                    else:
                        is_file = True
                        node = value
                        is_done = False
                        break

        raise Exception("Invalid configuration for content.")



import sys
import setuptools
import re

install_requires = list()
dependency_links = list()
uninstalled_pkgs = list()
with open("requirements.txt", "r", encoding='utf-8') as f:
    for line in f:
        if re.match("^[\w_-]*(==(\d+\.)*\d+){0,1}$", line.strip()):
            install_requires.append(line.strip())
        elif line.strip() and not line.strip().startswith("#"):
            uninstalled_pkgs.append(line.strip())

for i, pkg in enumerate(uninstalled_pkgs):
    if pkg.startswith("-e "):
        dependency_links.append(pkg.lstrip("-e").strip())
        del uninstalled_pkgs[i]

setuptools.setup(name='notest',
      version='0.0.6',
      description='Not Only Test! One Excellent Python Testing Tool',
      long_description='Not Only Test! One Excellent Python Testing Tool.',
      author='Chuanhao Qu',
      author_email='quchuanhao@gmail.com',
      url='https://github.com/godq/pyresttest',
      keywords=['rest', 'web', 'http', 'testing', "notest", "api"],
      classifiers=[
          'Environment :: Console',
          'License :: OSI Approved :: Apache Software License',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Software Development :: Testing',
          'Topic :: Software Development :: Quality Assurance',
          'Topic :: Utilities'
      ],
      packages=setuptools.find_packages(where='.', exclude=(), include=('notest*')),
      license='Apache License, Version 2.0',
      install_requires=install_requires,
      dependency_links=dependency_links,
      include_package_data=True,
      zip_safe=True,
      entry_points={
        'console_scripts': [
            'notest = notest.main:command_line_run',
        ],
      },
)



import json


class TestResult:
    """ Encapsulates everything about a test response """
    test_type = None
    test = None  # Test run
    passed = False
    failures = None
    loop = False

    def __init__(self):
        self.failures = list()

    def __str__(self):
        msg = list()
        msg.append("\n====================")
        msg.append("Test Type: {}".format(self.test.test_type))
        msg.append("Passed? : {}".format(self.passed))
        msg.append("Failures : {}".format(self.failures))
        msg.append("====================\n")

        return "\n".join(msg)


import traceback
import json
import yaml
import jsonschema

from notest import validators
from notest.lib import parsing
from notest import contenthandling


class JsonSchemaValidator(validators.AbstractValidator):
    """ Json schema validator using the jsonschema library """
    schema = None

    def validate(self, body=None, headers=None, context=None):
        schema_text = self.schema.get_content(context=context)
        schema = yaml.safe_load(schema_text)
        # TODO add caching of parsed schema

        try:
            # TODO try draft3/draft4 iter_errors -
            # https://python-jsonschema.readthedocs.org/en/latest/validate/#jsonschema.IValidator.iter_errors
            parsed_body = body
            if isinstance(body, bytes):
                parsed_body = body.decode()
            jsonschema.validate(json.loads(parsed_body), schema)
            return True
        except jsonschema.exceptions.ValidationError as ve:
            trace = traceback.format_exc()
            return validators.Failure(
                message="JSON Schema Validation Failed", 
                details=trace,
                validator=self,
                failure_type=validators.FAILURE_VALIDATOR_EXCEPTION)

    def get_readable_config(self, context=None):
        return "JSON schema validation"

    @classmethod
    def parse(cls, config):
        validator = JsonSchemaValidator()
        config = parsing.lowercase_keys(config)
        if 'schema' not in config:
            raise ValueError(
                "Cannot create schema validator without a 'schema' configuration element!")
        validator.schema = contenthandling.ContentHandler.parse_content(
            config[
                'schema'])
        return validator


VALIDATORS = {'json_schema': JsonSchemaValidator.parse}
